#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-06-20
# @Filename: mwm_cb_uvex.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee
from peewee import fn

from sdssdb.peewee.sdss5db.catalogdb import (UVOT_SSC_1, XMM_OM_SUSS_4_1,
                                             Catalog, CatalogToGUVCat,
                                             CatalogToTIC_v8,
                                             CatalogToUVOT_SSC_1,
                                             CatalogToXMM_OM_SUSS_4_1,
                                             Gaia_DR2,
                                             GeometricDistances_Gaia_DR2,
                                             GUVCat, TIC_v8, TwoMassPSC)

from . import BaseCarton


fuv_mag = GUVCat.fuv_mag
nuv_mag = GUVCat.nuv_mag
fuv_magerr = GUVCat.fuv_magerr
nuv_magerr = GUVCat.nuv_magerr
fuv_nuv = fuv_mag - nuv_mag

BJ = GeometricDistances_Gaia_DR2

absg = (Gaia_DR2.phot_g_mean_mag - 5 * fn.log(BJ.r_est) + 5)
bp_rp = Gaia_DR2.bp_rp

gmagab = Gaia_DR2.phot_g_mean_mag - 25.688 + 25.792
fuvg = fuv_mag - gmagab

B_AB = Gaia_DR2.phot_bp_mean_mag - 25.351 + 25.386
R_AB = Gaia_DR2.phot_rp_mean_mag - 24.762 + 25.116
B_AB_R_AB = B_AB - R_AB

colour_absg = ((absg > 4.09) | (absg > 4.5457 * bp_rp + 4.0457))

pmra = Gaia_DR2.pmra
pmdec = Gaia_DR2.pmdec
pmra_error = Gaia_DR2.pmra_error
pmdec_error = Gaia_DR2.pmdec_error

pm = fn.sqrt(pmra * pmra + pmdec * pmdec)
pm_error = fn.sqrt(fn.pow((pmra_error * pmra / pm), 2) +
                   fn.pow((pmdec_error * pmdec / pm), 2))
logpmdpm = fn.log(pm / pm_error)

epar = Gaia_DR2.parallax / Gaia_DR2.parallax_error

astrometric_cuts = (BJ.r_lo <= 1500,
                    ~(((logpmdpm < 0.301) &
                       (epar > -1.4995 * logpmdpm - 4.05) &
                       (epar < 1.4995 * logpmdpm + 4.05)) |
                      (fn.pow((logpmdpm - 0.301), 2) / (0.39794 * 0.39794) +
                       fn.pow(epar, 2) / (4.5 * 4.5) <= 1)))


class MWM_CB_UVEX1_Carton(BaseCarton):
    """MWM Compact Binaries UV excess 1.

    Definition:

       Match Gaia - GALEX, use both bands of GALEX (FUV and NUV),
       apply color cuts to select objects with UV excess between main
       sequence and WD sequence. Remove objects with low parallax and
       proper motion accuracy to not get swamped by AGN and QSO.

    Pseudo-SQL:

        - Match Gaia DR2 with GALEX within 5 arcsec, keep the nearest match
          only, add distance and lower limit distance (columns r_est, r_lo;
          estimated and lower limit to estimated distance) from catalog
          gdr2_contrib/geometric_distance; use columns fuv_mag, nuv_mag,
          fuv_magerr, nuv_magerr from GALEX catalog.

        - Definitions:
            - pm = sqrt(pmra*pmra+pmdec*pmdec)
            - pm_error = sqrt((pmra_error*pmra/pm)*(pmra_error*pmra/pm) +
                         (pmdec_error*pmdec/pm)*(pmdec_error*pmdec/pm))
            - logpmdpm = log10(pm/pm_error)
            - gmagab = phot_g_mean_mag-25.688+25.792
            - bmagab = phot_bp_mean_mag-25.351+25.386
            - rmagab = phot_rp_mean_mag-24.762+25.116
            - B_AB-R_AB = bmagab-rmagab
            - absg=phot_g_mean_mag-5*log10(r_est)+5
            - nuvg=nuv_mag-gmagab
            - fuvg=fuv_mag-gmagab
            - fuvnub=fuv_mag-nuv_mag
            - bp_rp=phot_bp_mean_mag-phot_rp_mean_mag

        - General cut:
            - Gaia: visibility_periods_used >5

        - Colour and magnitude cuts:
            - nuv_magerr < 0.2 && nuv_mag>-100 &&
                (fuv_magerr >=0.2 || fuv_mag< -100)
            - absg < (absg > 4.09 || absg > 4.5457*bp_rp + 4.0457)  [colour_abs]
            - ((nuv_mag-gmagab) < 2.25 ||
                ((nuv_mag-gmagab) < 6.725*(B_AB-R_AB)-1.735  &&
                 (nuv_mag-gmagab)< -0.983*(B_AB-R_AB)+8.24))

        - Astrometric cuts:
            - !((logpmdpm <0.301 &&
                (parallax/parallax_error)>-1.4995*logpmdpm-4.05 &&
                (parallax/parallax_error)<1.4995*logpmdpm+4.05) ||
               (logpmdpm-0.301)*(logpmdpm-0.301)/(0.39794*0.39794) +
               (parallax/parallax_error)^2/(4.5*4.5)<=1)
            - r_lo <= 1500

    """

    name = 'mwm_cb_uvex1'
    mapper = 'MWM'
    category = 'science'
    program = 'mwm_cb'
    cadence = 'mwm_cb_2x1'
    priority = 1400

    def setup_transaction(self):

        self.database.execute_sql('SET LOCAL join_collapse_limit = 1;')
        super().setup_transaction()

    def build_query(self, version_id, query_region=None):

        colour_cuts = (fuv_magerr < 0.2,
                       nuv_magerr < 0.2,
                       fuv_mag > -100,
                       nuv_mag > -100,
                       colour_absg,
                       (absg > -1.11749253e-03 * fn.pow(fuv_nuv, 3) +
                        1.53748615e-02 * fn.pow(fuv_nuv, 2) +
                        3.66419895e-01 * fuv_nuv + 2.20026639),
                       ((fuvg < 6.08) | ((fuvg < 11.82 * B_AB_R_AB + 2.58) &
                                         (fuvg < -0.79 * B_AB_R_AB + 9.21))))

        query = (GUVCat
                 .select(CatalogToTIC_v8.catalogid,
                         Gaia_DR2.source_id,
                         Gaia_DR2.ra.alias('gaia_ra'),
                         Gaia_DR2.dec.alias('gaia_dec'),
                         Gaia_DR2.pmra,
                         Gaia_DR2.pmdec,
                         Gaia_DR2.pmra_error,
                         Gaia_DR2.pmdec_error,
                         Gaia_DR2.parallax,
                         Gaia_DR2.parallax_error,
                         Gaia_DR2.phot_g_mean_mag,
                         Gaia_DR2.phot_bp_mean_mag,
                         Gaia_DR2.phot_rp_mean_mag,
                         BJ.r_est,
                         BJ.r_lo,
                         GUVCat.objid.alias('guvcat_objid'),
                         GUVCat.ra.alias('guvcat_ra'),
                         GUVCat.dec.alias('guvcat_dec'),
                         GUVCat.nuv_mag,
                         GUVCat.fuv_mag,
                         GUVCat.nuv_magerr,
                         GUVCat.fuv_magerr)
                 .join(CatalogToGUVCat)
                 .join(CatalogToTIC_v8, on=(CatalogToGUVCat.catalogid ==
                                            CatalogToTIC_v8.catalogid))
                 .join(TIC_v8)
                 .join(Gaia_DR2)
                 .join(BJ)
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True)
                 .where(CatalogToGUVCat.version_id == version_id,
                        CatalogToGUVCat.best >> True)
                 .where(Gaia_DR2.visibility_periods_used > 5)
                 .where(*colour_cuts)
                 .where(*astrometric_cuts))

        if query_region:
            query = (query
                     .join_from(CatalogToTIC_v8, Catalog)
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))

        return query


class MWM_CB_UVEX2_Carton(BaseCarton):
    """MWM Compact Binaries UV excess 2.

    Definition:

        Match Gaia - GALEX, use NUV channel of GALEX only, apply color cuts
        to select objects with UV excess between main sequence and WD sequence.
        Remove objects with low parallax and proper motion accuracy to not
        get swamped by AGN and QSO.

    Pseudo-SQL:

        - Match Gaia DR2 with GALEX within 5 arcsec, keep the nearest match
          only, add distance and lower limit distance (columns r_est, r_lo;
          estimated and lower limit to estimated distance) from catalog
          gdr2_contrib/geometric_distance; use columns fuv_mag, nuv_mag,
          fuv_magerr, nuv_magerr from GALEX catalog.

        - Definitions:
            - pm = sqrt(pmra*pmra+pmdec*pmdec)
            - pm_error = sqrt((pmra_error*pmra/pm)*(pmra_error*pmra/pm) +
                         (pmdec_error*pmdec/pm)*(pmdec_error*pmdec/pm))
            - logpmdpm = log10(pm/pm_error)
            - gmagab = phot_g_mean_mag-25.688+25.792
            - bmagab = phot_bp_mean_mag-25.351+25.386
            - rmagab = phot_rp_mean_mag-24.762+25.116
            - B_AB-R_AB = bmagab-rmagab
            - absg=phot_g_mean_mag-5*log10(r_est)+5
            - nuvg=nuv_mag-gmagab
            - fuvg=fuv_mag-gmagab
            - fuvnub=fuv_mag-nuv_mag
            - bp_rp=phot_bp_mean_mag-phot_rp_mean_mag

        - General cut:
            - Gaia: visibility_periods_used >5

        - Colour and magnitude cuts:
            - fuv_magerr < 0.2 && nuv_magerr < 0.2 &&
              fuv_mag > -100 && nuv_mag > -100
            - absg < (absg > 4.09 || absg > 4.5457*bp_rp + 4.0457)  [colour_abs]
            - absg > -1.11749253e-03*fuv_nuv^3 +
                1.53748615e-02*fuv_nuv^2 +
                3.66419895e-01*fuv_nuv+2.20026639e+00
            - (fuvg < 6.08 || (fuvg < 11.82*(B_AB-R_AB) +
                               2.58 && fuvg < -0.79*(B_AB-R_AB)+9.21))

        - Astrometric cuts:
            - !((logpmdpm <0.301 &&
                (parallax/parallax_error)>-1.4995*logpmdpm-4.05 &&
                (parallax/parallax_error)<1.4995*logpmdpm+4.05) ||
               (logpmdpm-0.301)*(logpmdpm-0.301)/(0.39794*0.39794) +
               (parallax/parallax_error)^2/(4.5*4.5)<=1)
            - r_lo <= 1500

    """

    name = 'mwm_cb_uvex2'
    mapper = 'MWM'
    category = 'science'
    program = 'mwm_cb'
    cadence = 'mwm_cb_2x1'
    priority = 1400

    def setup_transaction(self):

        self.database.execute_sql('SET LOCAL join_collapse_limit = 1;')
        super().setup_transaction()

    def build_query(self, version_id, query_region=None):

        colour_cuts = (nuv_magerr < 0.2,
                       nuv_mag > -100,
                       colour_absg,
                       (((nuv_mag - gmagab) < 2.25) |
                        (((nuv_mag - gmagab) < 6.725 * B_AB_R_AB - 1.735) &
                         ((nuv_mag - gmagab) < -0.983 * B_AB_R_AB + 8.24))))

        query = (GUVCat
                 .select(CatalogToGUVCat.catalogid,
                         Gaia_DR2.source_id,
                         Gaia_DR2.ra.alias('gaia_ra'),
                         Gaia_DR2.dec.alias('gaia_dec'),
                         Gaia_DR2.pmra,
                         Gaia_DR2.pmdec,
                         Gaia_DR2.pmra_error,
                         Gaia_DR2.pmdec_error,
                         Gaia_DR2.parallax,
                         Gaia_DR2.parallax_error,
                         Gaia_DR2.phot_g_mean_mag,
                         Gaia_DR2.phot_bp_mean_mag,
                         Gaia_DR2.phot_rp_mean_mag,
                         BJ.r_est,
                         BJ.r_lo,
                         GUVCat.objid.alias('guvcat_objid'),
                         GUVCat.ra.alias('guvcat_ra'),
                         GUVCat.dec.alias('guvcat_dec'),
                         GUVCat.nuv_mag,
                         GUVCat.nuv_magerr)
                 .join(CatalogToGUVCat)
                 .join(CatalogToTIC_v8, on=(CatalogToGUVCat.catalogid ==
                                            CatalogToTIC_v8.catalogid))
                 .join(TIC_v8)
                 .join(Gaia_DR2)
                 .join(BJ)
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True)
                 .where(CatalogToGUVCat.version_id == version_id,
                        CatalogToGUVCat.best >> True)
                 .where(Gaia_DR2.visibility_periods_used > 5)
                 .where(*colour_cuts)
                 .where(*astrometric_cuts))

        if query_region:
            query = (query
                     .join_from(CatalogToTIC_v8, Catalog)
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))

        return query


class MWM_CB_UVEX3_Carton(BaseCarton):
    """MWM Compact Binaries UV excess 3.

    Definition:

       Match Gaia - XMMOMSUOB (or OM SUSS v4.1), apply color cuts to select
       objects with UV excess between main sequence and WD sequence.
       Remove objects with low parallax and proper motion accuracy
       to not get swamped by AGN and QSO.

    Pseudo-SQL:

        - Apply some general UV quality flags and work on subset of entries
          in OM-SUSS:
            - (UVW2_AB_MAG >-100) || (UVW1_AB_MAG >-100) || (UVM2_AB_MAG>-100)
            - !(equals(substring(UVW2_QUALITY_FLAG_ST,0,1),"T") ||
                equals(substring(UVM2_QUALITY_FLAG_ST,0,1),"T") ||
                equals(substring(UVW1_QUALITY_FLAG_ST,0,1),"T") ||
                (equals(substring(UVW2_QUALITY_FLAG_ST,1,2),"T") &&
                 UVW2_SIGNIF <10) ||
                (equals(substring(UVM2_QUALITY_FLAG_ST,1,2),"T") &&
                 UVM2_SIGNIF <10) ||
                (equals(substring(UVW1_QUALITY_FLAG_ST,1,2),"T") &&
                 UVW1_SIGNIF <10) ||
                (equals(substring(UVW2_QUALITY_FLAG_ST,2,3) ,"T") &&
                 UVW2_SIGNIF <10) ||
                (equals(substring(UVM2_QUALITY_FLAG_ST,2,3),"T") &&
                 UVM2_SIGNIF <10) ||
                (equals(substring(UVW1_QUALITY_FLAG_ST,2,3),"T") &&
                 UVW1_SIGNIF <10) ||
                (equals(substring(UVW2_QUALITY_FLAG_ST,6,7) ,"T")) ||
                equals(substring(UVM2_QUALITY_FLAG_ST,6,7),"T") ||
                equals(substring(UVW1_QUALITY_FLAG_ST,6,7),"T") ||
                equals(substring(UVW2_QUALITY_FLAG_ST,7,8),"T") ||
                equals(substring(UVM2_QUALITY_FLAG_ST,7,8),"T") ||
                equals(substring(UVW1_QUALITY_FLAG_ST,7,8),"T") ||
                equals(substring(UVW2_QUALITY_FLAG_ST,8,9),"T") ||
                equals(substring(UVM2_QUALITY_FLAG_ST,8,9),"T") ||
                equals(substring(UVW1_QUALITY_FLAG_ST,8,9),"T"))

        - Match UV catalog with Gaia DR2 within 3 arcsec, keep the nearest
          match only, add distance and lower limit distance (columns r_est,
          r_lo; estimated and lower limit to estimated distance)
          from catalog gdr2_contrib/geometric_distance.

        - Definitions:
            - pm = sqrt(pmra*pmra+pmdec*pmdec)
            - pm_error = sqrt((pmra_error*pmra/pm)*(pmra_error*pmra/pm) +
                         (pmdec_error*pmdec/pm)*(pmdec_error*pmdec/pm))
            - logpmdpm = log10(pm/pm_error)
            - gmagab = phot_g_mean_mag-25.688+25.792
            - bmagab = phot_bp_mean_mag-25.351+25.386
            - rmagab = phot_rp_mean_mag-24.762+25.116
            - B_AB-R_AB = bmagab-rmagab
            - absg=phot_g_mean_mag-5*log10(r_est)+5
            - nuvg=nuv_mag-gmagab
            - fuvg=fuv_mag-gmagab
            - fuvnub=fuv_mag-nuv_mag
            - bp_rp=phot_bp_mean_mag-phot_rp_mean_mag

        - General cut:
            - Gaia: visibility_periods_used >5

        - Colour and magnitude cuts:
            - absg < (absg > 4.09 || absg > 4.5457*bp_rp + 4.0457)  [colour_abs]
            - ((UVM2_AB_MAG - gmagab) < 2.25 ||
               ((UVM2_AB_MAG - gmagab) < 6 &&
                (UVM2_AB_MAG - gmagab) < 5.57377*(B_AB-R_AB)+0.2049))

        - Astrometric cuts:
            - !((logpmdpm <0.301 &&
                (parallax/parallax_error)>-1.4995*logpmdpm-4.05 &&
                (parallax/parallax_error)<1.4995*logpmdpm+4.05) ||
               (logpmdpm-0.301)*(logpmdpm-0.301)/(0.39794*0.39794) +
               (parallax/parallax_error)^2/(4.5*4.5)<=1)
            - r_lo <= 1500

    """

    name = 'mwm_cb_uvex3'
    mapper = 'MWM'
    category = 'science'
    program = 'mwm_cb'
    cadence = 'mwm_cb_2x1'
    priority = 1400

    def build_query(self, version_id, query_region=None):

        uvw2_ab_mag = XMM_OM_SUSS_4_1.uvw2_ab_mag
        uvw1_ab_mag = XMM_OM_SUSS_4_1.uvw1_ab_mag
        uvm2_ab_mag = XMM_OM_SUSS_4_1.uvm2_ab_mag

        uvw2_quality_flag_st = XMM_OM_SUSS_4_1.uvw2_quality_flag_st
        uvm2_quality_flag_st = XMM_OM_SUSS_4_1.uvm2_quality_flag_st
        uvw1_quality_flag_st = XMM_OM_SUSS_4_1.uvw1_quality_flag_st

        uvw2_signif = XMM_OM_SUSS_4_1.uvw2_signif
        uvm2_signif = XMM_OM_SUSS_4_1.uvm2_signif
        uvw1_signif = XMM_OM_SUSS_4_1.uvw1_signif

        quality_flags = (
            (uvw2_ab_mag > -100) | (uvw1_ab_mag > -100) | (uvm2_ab_mag > -100),
            ~((fn.substr(uvw2_quality_flag_st, 1, 1) == 'T') |
              (fn.substr(uvm2_quality_flag_st, 1, 1) == 'T') |
              (fn.substr(uvw1_quality_flag_st, 1, 1) == 'T') |
              ((fn.substr(uvw2_quality_flag_st, 2, 1) == 'T') & (uvw2_signif < 10)) |
              ((fn.substr(uvm2_quality_flag_st, 2, 1) == 'T') & (uvm2_signif < 10)) |
              ((fn.substr(uvw1_quality_flag_st, 2, 1) == 'T') & (uvw1_signif < 10)) |
              ((fn.substr(uvw2_quality_flag_st, 3, 1) == 'T') & (uvw2_signif < 10)) |
              ((fn.substr(uvm2_quality_flag_st, 3, 1) == 'T') & (uvm2_signif < 10)) |
              ((fn.substr(uvw1_quality_flag_st, 3, 1) == 'T') & (uvw1_signif < 10)) |
              ((fn.substr(uvw2_quality_flag_st, 7, 1) == 'T') |
               (fn.substr(uvm2_quality_flag_st, 7, 1) == 'T') |
               (fn.substr(uvw1_quality_flag_st, 7, 1) == 'T') |
               (fn.substr(uvw2_quality_flag_st, 8, 1) == 'T') |
               (fn.substr(uvm2_quality_flag_st, 8, 1) == 'T') |
               (fn.substr(uvw1_quality_flag_st, 8, 1) == 'T') |
               (fn.substr(uvw2_quality_flag_st, 9, 1) == 'T') |
               (fn.substr(uvm2_quality_flag_st, 9, 1) == 'T') |
               (fn.substr(uvw1_quality_flag_st, 9, 1) == 'T')))
        )

        quality_cte = (XMM_OM_SUSS_4_1
                       .select(XMM_OM_SUSS_4_1,
                               CatalogToXMM_OM_SUSS_4_1.catalogid)
                       .join(CatalogToXMM_OM_SUSS_4_1)
                       .where(CatalogToXMM_OM_SUSS_4_1.version_id == version_id,
                              CatalogToXMM_OM_SUSS_4_1.best >> True)
                       .where(*quality_flags)
                       .cte('quality_cte', materialized=True))

        colour_cuts = (colour_absg,
                       ((quality_cte.c.uvm2_ab_mag - gmagab) < 2.25) |
                       (((quality_cte.c.uvm2_ab_mag - gmagab) < 6) &
                        ((quality_cte.c.uvm2_ab_mag - gmagab) < 5.57377 * (B_AB_R_AB) + 0.2049)))

        query = (Catalog
                 .select(Catalog.catalogid,
                         Gaia_DR2.source_id,
                         Gaia_DR2.ra.alias('gaia_ra'),
                         Gaia_DR2.dec.alias('gaia_dec'),
                         Gaia_DR2.pmra,
                         Gaia_DR2.pmdec,
                         Gaia_DR2.pmra_error,
                         Gaia_DR2.pmdec_error,
                         Gaia_DR2.parallax,
                         Gaia_DR2.parallax_error,
                         Gaia_DR2.phot_g_mean_mag,
                         Gaia_DR2.phot_bp_mean_mag,
                         Gaia_DR2.phot_rp_mean_mag,
                         gmagab.alias('gmagab'),
                         BJ.r_est,
                         BJ.r_lo,
                         XMM_OM_SUSS_4_1.srcnum,
                         quality_cte.c.uvm2_ab_mag.alias('uvm2_ab_mag'),
                         quality_cte.c.uvw2_ab_mag.alias('uvw2_ab_mag'),
                         quality_cte.c.uvw1_ab_mag.alias('uvw1_ab_mag'),
                         quality_cte.c.uvm2_quality_flag_st.alias('uvm2_quality_flag_st'),
                         quality_cte.c.uvw2_quality_flag_st.alias('uvw2_quality_flag_st'),
                         quality_cte.c.uvw1_quality_flag_st.alias('uvw1_quality_flag_st'))
                 .join(CatalogToTIC_v8)
                 .join(TIC_v8)
                 .join(Gaia_DR2)
                 .join(BJ)
                 .join_from(Catalog, quality_cte,
                            on=(quality_cte.c.catalogid == Catalog.catalogid))
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True)
                 .where(Gaia_DR2.visibility_periods_used > 5)
                 .where(*colour_cuts)
                 .where(*astrometric_cuts)
                 .with_cte(quality_cte))

        if query_region:
            query = (query
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))

        return query


class MWM_CB_UVEX4_Carton(BaseCarton):
    """MWM Compact Binaries UV excess 4.

    Definition:

       Match Gaia - SWUVOTSSC (Swift UVOT catalog), apply color cuts to
       select objects with UV excess between main sequence and WD sequence.
       Remove objects with low parallax and proper motion accuracy to not
       get swamped by AGN and QSO.

    Pseudo-SQL:

        - Apply some general UV quality flags and work on subset of entries
        in UVOT catalog::
            - (UVW2_AB_MAG >-100) || (UVW1_AB_MAG >-100) || (UVM2_AB_MAG>-100)
            - !(equals(substring(UVW1_QUALITY_FLAG_BIN,8,9),"1") ||
                equals(substring(UVW2_QUALITY_FLAG_BIN,8,9),"1") ||
                equals(substring(UVM2_QUALITY_FLAG_BIN,8,9),"1") ||
                (equals(substring(UVW1_QUALITY_FLAG_BIN,7,8),"1") && UVW1_SIGNIF <10) ||
                (equals(substring(UVW2_QUALITY_FLAG_BIN,7,8),"1") && UVW2_SIGNIF <10) ||
                (equals(substring(UVM2_QUALITY_FLAG_BIN,7,8),"1") && UVM2_SIGNIF <10) ||
                (equals(substring(UVW1_QUALITY_FLAG_BIN,6,7),"1") && UVW1_SIGNIF <10) ||
                (equals(substring(UVW2_QUALITY_FLAG_BIN,6,7),"1") && UVW2_SIGNIF <10) ||
                (equals(substring(UVM2_QUALITY_FLAG_BIN,6,7),"1") && UVM2_SIGNIF <10) ||
                (equals(substring(UVW1_QUALITY_FLAG_BIN,3,4),"1") && UVW1_SIGNIF <10) ||
                (equals(substring(UVW2_QUALITY_FLAG_BIN,3,4),"1") && UVW2_SIGNIF <10) ||
                (equals(substring(UVM2_QUALITY_FLAG_BIN,3,4),"1") && UVM2_SIGNIF <10) ||
                equals(substring(UVW1_QUALITY_FLAG_BIN,2,3),"1") ||
                equals(substring(UVW2_QUALITY_FLAG_BIN,2,3),"1") ||
                equals(substring(UVM2_QUALITY_FLAG_BIN,2,3),"1"))

        - Match UV catalog with Gaia DR2 within 3 arcsec, keep the nearest
          match only, add distance and lower limit distance (columns r_est,
          r_lo; estimated and lower limit to estimated distance)
          from catalog gdr2_contrib/geometric_distance.

        - Definitions:
            - pm = sqrt(pmra*pmra+pmdec*pmdec)
            - pm_error = sqrt((pmra_error*pmra/pm)*(pmra_error*pmra/pm) +
                         (pmdec_error*pmdec/pm)*(pmdec_error*pmdec/pm))
            - logpmdpm = log10(pm/pm_error)
            - gmagab = phot_g_mean_mag-25.688+25.792
            - bmagab = phot_bp_mean_mag-25.351+25.386
            - rmagab = phot_rp_mean_mag-24.762+25.116
            - B_AB-R_AB = bmagab-rmagab
            - absg=phot_g_mean_mag-5*log10(r_est)+5
            - nuvg=nuv_mag-gmagab
            - fuvg=fuv_mag-gmagab
            - fuvnub=fuv_mag-nuv_mag
            - bp_rp=phot_bp_mean_mag-phot_rp_mean_mag

        - General cut:
            - Gaia: visibility_periods_used >5

        - Colour and magnitude cuts:
            - absg < (absg > 4.09 || absg > 4.5457*bp_rp + 4.0457)  [colour_abs]
            - ((UVM2_AB_MAG - gmagab) < 2.25 ||
               ((UVM2_AB_MAG - gmagab) < 6 &&
                (UVM2_AB_MAG - gmagab) < 5.57377*(B_AB-R_AB)+0.2049))

        - Astrometric cuts:
            - !((logpmdpm <0.301 &&
                (parallax/parallax_error)>-1.4995*logpmdpm-4.05 &&
                (parallax/parallax_error)<1.4995*logpmdpm+4.05) ||
               (logpmdpm-0.301)*(logpmdpm-0.301)/(0.39794*0.39794) +
               (parallax/parallax_error)^2/(4.5*4.5)<=1)
            - r_lo <= 1500

    """

    name = 'mwm_cb_uvex4'
    mapper = 'MWM'
    category = 'science'
    program = 'mwm_cb'
    cadence = 'mwm_cb_2x1'
    priority = 1400

    def build_query(self, version_id, query_region=None):

        uvw2_ab_mag = UVOT_SSC_1.uvw2_ab
        uvm2_ab_mag = UVOT_SSC_1.uvm2_ab
        uvw1_ab_mag = UVOT_SSC_1.uvw1_ab

        uvw2_quality_flag = UVOT_SSC_1.fuvw2
        uvm2_quality_flag = UVOT_SSC_1.fuvm2
        uvw1_quality_flag = UVOT_SSC_1.fuvw1

        uvw2_signif = UVOT_SSC_1.suvw2
        uvm2_signif = UVOT_SSC_1.suvm2
        uvw1_signif = UVOT_SSC_1.suvw1

        # uvXY_quality_flag is a 9-bit integer so character 8 (0-indexed) in
        # the binary string is the first bit, 7 is the second, etc.

        quality_flags = (
            (uvw2_ab_mag > -100) | (uvw1_ab_mag > -100) | (uvm2_ab_mag > -100),
            ~((uvw1_quality_flag.bin_and(2**0) > 0) |
              (uvw2_quality_flag.bin_and(2**0) > 0) |
              (uvm2_quality_flag.bin_and(2**0) > 0) |
              (uvw1_quality_flag.bin_and(2**1) > 0) & (uvw1_signif < 10) |
              (uvw2_quality_flag.bin_and(2**1) > 0) & (uvw2_signif < 10) |
              (uvm2_quality_flag.bin_and(2**1) > 0) & (uvm2_signif < 10) |
              (uvw1_quality_flag.bin_and(2**2) > 0) & (uvw1_signif < 10) |
              (uvw2_quality_flag.bin_and(2**2) > 0) & (uvw2_signif < 10) |
              (uvm2_quality_flag.bin_and(2**2) > 0) & (uvm2_signif < 10) |
              (uvw1_quality_flag.bin_and(2**5) > 0) & (uvw1_signif < 10) |
              (uvw2_quality_flag.bin_and(2**5) > 0) & (uvw2_signif < 10) |
              (uvm2_quality_flag.bin_and(2**5) > 0) & (uvm2_signif < 10) |
              (uvw1_quality_flag.bin_and(2**6) > 0) |
              (uvw2_quality_flag.bin_and(2**6) > 0) |
              (uvm2_quality_flag.bin_and(2**6) > 0))
        )

        quality_cte = (UVOT_SSC_1
                       .select(UVOT_SSC_1,
                               CatalogToUVOT_SSC_1.catalogid)
                       .join(CatalogToUVOT_SSC_1)
                       .where(CatalogToUVOT_SSC_1.version_id == version_id,
                              CatalogToUVOT_SSC_1.best >> True)
                       .where(*quality_flags)
                       .cte('quality_cte', materialized=True))

        colour_cuts = (colour_absg,
                       ((quality_cte.c.uvm2_ab - gmagab) < 2.25) |
                       (((quality_cte.c.uvm2_ab - gmagab) < 6) &
                        ((quality_cte.c.uvm2_ab - gmagab) < 5.57377 * (B_AB_R_AB) + 0.2049)))

        query = (Catalog
                 .select(Catalog.catalogid,
                         Gaia_DR2.source_id,
                         Gaia_DR2.ra.alias('gaia_ra'),
                         Gaia_DR2.dec.alias('gaia_dec'),
                         Gaia_DR2.pmra,
                         Gaia_DR2.pmdec,
                         Gaia_DR2.pmra_error,
                         Gaia_DR2.pmdec_error,
                         Gaia_DR2.parallax,
                         Gaia_DR2.parallax_error,
                         Gaia_DR2.phot_g_mean_mag,
                         Gaia_DR2.phot_bp_mean_mag,
                         Gaia_DR2.phot_rp_mean_mag,
                         gmagab.alias('gmagab'),
                         BJ.r_est,
                         BJ.r_lo,
                         UVOT_SSC_1.srcnum,
                         quality_cte.c.uvm2_ab.alias('uvm2_ab'),
                         quality_cte.c.uvw2_ab.alias('uvw2_ab'),
                         quality_cte.c.uvw1_ab.alias('uvw1_ab'),
                         quality_cte.c.fuvw2.alias('fuvw2'),
                         quality_cte.c.fuvm2.alias('fuvm2'),
                         quality_cte.c.fuvw1.alias('fuvw1'))
                 .join(CatalogToTIC_v8)
                 .join(TIC_v8)
                 .join(Gaia_DR2)
                 .join(BJ)
                 .join_from(Catalog, quality_cte,
                            on=(quality_cte.c.catalogid == Catalog.catalogid))
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True)
                 .where(Gaia_DR2.visibility_periods_used > 5)
                 .where(*colour_cuts)
                 .where(*astrometric_cuts)
                 .with_cte(quality_cte))

        if query_region:
            query = (query
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))

        return query


class MWM_CB_UVEX5_Carton(BaseCarton):
    """MWM Compact Binaries UV excess 5.

    Definition:

       Match Gaia - GALEX, use both bands of GALEX (FUV and NUV), apply colour
       cuts to select objects with UV excess between main sequence and WD
       sequence. Remove objects with low parallax and proper motion accuracy to
       not get swamped by AGN and QSO.

    Pseudo-SQL:

        - Match Gaia DR2 with GALEX within 5 arcsec, keep the nearest match
          only, add distance and lower limit distance (columns r_est, r_lo;
          estimated and lower limit to estimated distance) from catalog
          gdr2_contrib/geometric_distance; use columns fuv_mag, nuv_mag,
          fuv_magerr, nuv_magerr from GALEX catalog.

        - Match with 2MASS.

        - Definitions:
            - pm = sqrt(pmra*pmra+pmdec*pmdec)
            - pm_error = sqrt((pmra_error*pmra/pm)*(pmra_error*pmra/pm) +
                         (pmdec_error*pmdec/pm)*(pmdec_error*pmdec/pm))
            - logpmdpm = log10(pm/pm_error)
            - gmr = phot_g_mean_mag - phot_rp_mean_mag
            - absg = phot_g_mean_mag-5*log10(r_est)+5
            - Main-sequence defined as polynomial:
                GMS = 0.00206868742*gmr**6 + 0.0401594518*gmr**5 -
                0.842512410*gmr**4 + 4.89384979*gmr**3 -
                12.3826637*gmr**2 + 17.0197205*gmr - 3.19987835

        - General cut:
            - Gaia: visibility_periods_used > 5

        - Colour and magnitude cuts:
            - Hmag < 15
            - fuv_magerr < 0.2 && nuv_magerr < 0.2 &&
              fuv_mag > -100 && nuv_mag > -100
            - abs(GMS-absg) <=0.5 && absg >=4.0866
            - r_est < 0.51 * pow(10, 0.2291*phot_g_mean_mag)

        - Astrometric cuts:
            - !((logpmdpm <0.301 &&
                (parallax/parallax_error)>-1.4995*logpmdpm-4.05 &&
                (parallax/parallax_error)<1.4995*logpmdpm+4.05) ||
               (logpmdpm-0.301)*(logpmdpm-0.301)/(0.39794*0.39794) +
               (parallax/parallax_error)^2/(4.5*4.5)<=1)
            - r_lo <= 1500

        This sequence yields 10,766 objects whose UV emission is though
        to arise from an unseen compact companion.

    """

    name = 'mwm_cb_uvex5'
    mapper = 'MWM'
    category = 'science'
    program = 'mwm_cb'
    cadence = 'mwm_cb_2x1'
    priority = 1400

    def setup_transaction(self):

        self.database.execute_sql('SET LOCAL join_collapse_limit = 1;')
        super().setup_transaction()

    def build_query(self, version_id, query_region=None):

        bmr = Gaia_DR2.phot_bp_mean_mag - Gaia_DR2.phot_rp_mean_mag
        GMS = (0.00206868742 * fn.pow(bmr, 6) + 0.0401594518 * fn.pow(bmr, 5) -
               0.842512410 * fn.pow(bmr, 4) + 4.89384979 * fn.pow(bmr, 3) -
               12.3826637 * fn.pow(bmr, 2) + 17.0197205 * bmr - 3.19987835)

        colour_cuts = (TwoMassPSC.h_m < 15,
                       fn.abs(GMS - absg) <= 0.5,
                       absg >= 4.0866,
                       BJ.r_est < 0.51 * fn.pow(10, 0.2291 * Gaia_DR2.phot_g_mean_mag))

        query = (GUVCat
                 .select(CatalogToTIC_v8.catalogid,
                         Gaia_DR2.source_id,
                         Gaia_DR2.ra.alias('gaia_ra'),
                         Gaia_DR2.dec.alias('gaia_dec'),
                         Gaia_DR2.pmra,
                         Gaia_DR2.pmdec,
                         Gaia_DR2.pmra_error,
                         Gaia_DR2.pmdec_error,
                         Gaia_DR2.parallax,
                         Gaia_DR2.parallax_error,
                         Gaia_DR2.phot_g_mean_mag,
                         Gaia_DR2.phot_bp_mean_mag,
                         Gaia_DR2.phot_rp_mean_mag,
                         BJ.r_est,
                         BJ.r_lo,
                         GUVCat.objid.alias('guvcat_objid'),
                         GUVCat.ra.alias('guvcat_ra'),
                         GUVCat.dec.alias('guvcat_dec'),
                         GUVCat.nuv_mag,
                         GUVCat.fuv_mag,
                         GUVCat.nuv_magerr,
                         GUVCat.fuv_magerr,
                         GMS.alias('GMS'))
                 .join(CatalogToGUVCat)
                 .join(CatalogToTIC_v8,
                       on=(CatalogToGUVCat.catalogid == CatalogToTIC_v8.catalogid))
                 .join(TIC_v8)
                 .join(TwoMassPSC)
                 .join_from(TIC_v8, Gaia_DR2)
                 .join(BJ)
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True)
                 .where(CatalogToGUVCat.version_id == version_id,
                        CatalogToGUVCat.best >> True)
                 .where(Gaia_DR2.visibility_periods_used > 5)
                 .where(fuv_magerr < 0.2,
                        nuv_magerr < 0.2,
                        fuv_mag > -100,
                        nuv_mag > -100)
                 .where(*colour_cuts)
                 .where(*astrometric_cuts))

        if query_region:
            query = (query
                     .join_from(CatalogToTIC_v8, Catalog)
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))

        return query



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Tom Dwelly
# @Date: 2020-06-17
# @Filename: bhm_spiders_clusters.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

# flake8: noqa
# isort: skip_file

import peewee
from peewee import JOIN
from peewee import fn
import sdssdb

from target_selection.cartons.base import BaseCarton
from target_selection.mag_flux import psfmag_minus_fiber2mag, AB2nMgy


from sdssdb.peewee.sdss5db.catalogdb import (Catalog,
                                             CatalogToLegacy_Survey_DR8,
                                             Legacy_Survey_DR8,
                                             BHM_Spiders_Clusters_Superset,
                                             BHM_eFEDS_Veto,
                                             SDSS_DR16_SpecObj,
                                             )


## This provides the following BHM cartons:

# bhm_spiders_clusters-efeds-sdss-redmapper
# bhm_spiders_clusters-efeds-hsc-redmapper
# bhm_spiders_clusters-efeds-ls-redmapper
# bhm_spiders_clusters-efeds-erosita
# bhm_spiders_clusters-wide
# bhm_spiders_clusters-deep

# Details: Start here
# https://wiki.sdss.org/display/OPS/Defining+target+selection+and+cadence+algorithms


'''
Pseudo SQL (optional):

    all Cartons selections use the following pseudo-SQL:
        SELECT * from bhm_spiders_clusters_superset AS x
        INNER JOIN legacy_survey_dr8 AS ls ON  x.ls_id = ls.ls_id
        LEFT JOIN bhm_efeds_veto AS v ON q3c_join(x.opt_ra,x.opt_dec,v.plug_ra,v.plug_dec,1.0)
        LEFT JOIN sdss_specobj_dr16 AS sp ON q3c_join(x.opt_ra,x.opt_dec,so.plug_ra,so.plug_dec,1.0)
        WHERE x.ero_version = "efeds_c940_clus"
        AND WHERE x.ero_det_like > X.X
        AND WHERE (ls.fiberflux_r > D.D OR ls.fiberflux_z > E.E)    # faint limits
        AND WHERE ls.fiberflux_r < CCC.C                                               # bright limit
        AND WHERE (v.plug_ra = NULL  OR v.sn_median_all < 1.x  OR v.zwarning > 0  OR v.z_err > 2e-3  OR v.z_err <= 0.0)
        AND WHERE (so.plug_ra = NULL OR so.sn_median_all < 1.x OR so.zwarning > 0 OR so.z_err > 2e-3 OR so.z_err <= 0.0 OR so.z_err >so.z )
    Carton-specific pseudo-SQL based on targetting bits in superset catalogue:
        bhm_spiders_clusters_efeds_sdss_redmapper
            AND WHERE x.xmatch_flags & (2^2 + 2^6 + 2^10 + 2^12 + 2^17 + 2^19) !=0
        bhm_spiders_clusters_efeds_hsc_redmapper
            AND WHERE x.xmatch_flags & (2^5 + 2^11) !=0
        bhm_spiders_clusters_efeds_ls_redmapper
            AND WHERE x.xmatch_flags & (2^1 + 2^3 + 2^8 + 2^15) !=0
        bhm_spiders_clusters_efeds_erosita
            AND WHERE x.xmatch_flags & (2^2 + 2^6 + 2^10 + 2^12 + 2^17    +    2^19 + 2^5 + 2^11  +    2^1 + 2^3 + 2^8 + 2^15)  = 0
'''


class BhmSpidersClusEfedsCarton(BaseCarton):
    ''' Parent class that provides the underlying selections for all SPIDERS Clusters eFEDS cartons'''

    name = 'bhm_spiders_clusters-efeds-base'
    base_name = 'bhm_spiders_clusters-efeds-base'
    category = 'science'
    mapper = 'BHM'
    program = 'bhm_spiders'
    tile = False
    priority = None
    cadence = 'bhm_spiders_1x8'
    bitmask = None
    mask_sense = True

    def build_query(self, version_id, query_region=None):
        c = Catalog.alias()
        x = BHM_Spiders_Clusters_Superset.alias()
        ls = Legacy_Survey_DR8.alias()
        c2ls = CatalogToLegacy_Survey_DR8.alias()
        v = BHM_eFEDS_Veto.alias()
        s = SDSS_DR16_SpecObj.alias()

        base_parameters = self.config['parameters'].get(self.base_name, None)
        assert base_parameters is not None


        fiberflux_r_max = AB2nMgy(base_parameters['fibermag_r_min'])
        fiberflux_r_min = AB2nMgy(base_parameters['fibermag_r_max'])
        fiberflux_z_min = AB2nMgy(base_parameters['fibermag_z_max'])

        flux30 = AB2nMgy(30.0)

        value = peewee.Value(base_parameters.get('value', 1.0)).cast('float').alias('value')
        match_radius_spectro = base_parameters['spec_join_radius']/3600.0

        p_f = base_parameters['priority_floor']
        priority = peewee.Case(None,
                               (
                                   ((x.target_priority == 0), p_f+0), # BCGs
                               ),
                               p_f + x.target_priority + 10)     # Member galaxies

        # legacysurvey mags - derived from fiberfluxes - with limits to avoid divide by zero errors
        # notes on convertion from ls_fibermag to sdss_fiber2mag:
        # https://wiki.mpe.mpg.de/eRosita/EroAGN_eFEDS/SDSSIVSpecialPlates#Estimating_SDSS_fiber2mag_.2A_from_legacysurvey_photometry
        # for mostly point-like sources, the average offset is sdss_fiber2mag_[griz] = ls_fibermag_[griz] + 0.55 mag
        # for mostly galaxies, there is a magnitude and band-dependent shift
        # sdss_fiber2mag_g = ls_fibermag_g + 0.46 mag   flux_ratio = ~0.65
        # sdss_fiber2mag_r = ls_fibermag_r + 0.55 mag   flux_ratio = ~0.60
        # sdss_fiber2mag_i = ls_fibermag_i + 0.44 mag   flux_ratio = ~0.67
        # sdss_fiber2mag_z = ls_fibermag_z + 0.39 mag   flux_ratio = ~0.70
        flux_ratio = {'g' : 0.65, 'r' : 0.60, 'i' : 0.67, 'z' : 0.70 }
        # Then add the correction from sdss_fiber2mag to sdss_psfmag

        # Notes on converting from sdss_fiber2mag to sdss_psfmag
        # https://wiki.sdss.org/display/OPS/Contents+of+targetdb.magnitude#Contentsoftargetdb.magnitude-WhatmagnitudestoputintotheplPlugMapfilesforBOSSplatetargets?

        magnitude_g = (psfmag_minus_fiber2mag('g') +
                       22.5-2.5*fn.log10(fn.greatest(flux30,ls.fiberflux_g*flux_ratio['g']))).cast('float')
        magnitude_r = (psfmag_minus_fiber2mag('r') +
                       22.5-2.5*fn.log10(fn.greatest(flux30,ls.fiberflux_r*flux_ratio['r']))).cast('float')
        magnitude_z = (psfmag_minus_fiber2mag('z') +
                       22.5-2.5*fn.log10(fn.greatest(flux30,ls.fiberflux_z*flux_ratio['z']))).cast('float')
        # the simplest possible interpolation - TODO do this better
        magnitude_i = (psfmag_minus_fiber2mag('i') +
                       22.5-2.5*fn.log10(fn.greatest(flux30,
                                                     0.5*(ls.fiberflux_r+
                                                          ls.fiberflux_z)*flux_ratio['i']))).cast('float')



        pmra = peewee.Value(0.0).cast('float').alias('pmra')
        pmdec = peewee.Value(0.0).cast('float').alias('pmdec')
        parallax = peewee.Value(0.0).cast('float').alias('parallax')

        query = (
            c
            .select(c.catalogid,
                    priority.alias('priority'),
                    value,
                    pmra,
                    pmdec,
                    parallax,
                    magnitude_g.alias("g"),
                    magnitude_r.alias("r"),
                    magnitude_i.alias("i"),
                    magnitude_z.alias("z"),
            )
            .join(c2ls)
            .join(ls)
            .join(x)
            .join(v, JOIN.LEFT_OUTER,
                  on=(fn.q3c_join(c.ra,c.dec,
                                 v.plug_ra,v.plug_dec,
                                 match_radius_spectro) &
                      (v.sn_median[3] >= base_parameters['spec_sn_thresh']) &
                      (v.zwarning == 0) &
                      (v.z_err <= base_parameters['spec_z_err_thresh']) &
                      (v.z_err > 0.0)
                      )
                  )
            .join(s, JOIN.LEFT_OUTER,
                  on=(fn.q3c_join(c.ra,c.dec,
                                  s.ra,s.dec,
                                  match_radius_spectro) &
                      (s.snmedian_i >= base_parameters['spec_sn_thresh']) &
                      (s.zwarning == 0) &
                      (s.zerr <= base_parameters['spec_z_err_thresh']) &
                      (s.zerr > 0.0) &
                      (s.scienceprimary > 0)
                  )
            )
            .where(c.version_id == version_id,
                   c2ls.version_id == version_id,
                   c2ls.best == True)
            .where(
                (x.ero_version == base_parameters['ero_version'] ),
                (v.pk.is_null()),
                (s.specobjid.is_null()),
                (ls.fibertotflux_r < fiberflux_r_max),
                (
                    (ls.fiberflux_r >= fiberflux_r_min) |
                    (ls.fiberflux_z >= fiberflux_z_min)
                ),
                (x.ero_det_like > base_parameters['det_like_min']),
                (x.target_has_spec == 0),
            )
            .distinct([ls.ls_id])   # avoid duplicates - trust the ls_id
        )
        if self.mask_sense is True:
            query = query.where(x.xmatch_flags.bin_and(self.bitmask) != 0)
        else:
            query = query.where(x.xmatch_flags.bin_and(self.bitmask) == 0)

        return query


#-------BHM SPIDERS eFEDS Clusters SDSS RedMapper------ #

class BhmSpidersClusEfedsSdssRedmapperCarton(BhmSpidersClusEfedsCarton):
    '''
    Carton-specific pseudo-SQL based on targetting bits in superset catalogue:
            AND WHERE x.xmatch_flags & (2^2 + 2^6 + 2^10 + 2^12 + 2^17 + 2^19) !=0
    '''

    name = 'bhm_spiders_clusters-efeds-sdss-redmapper'
    bitmask = 2**2 + 2**6 + 2**10 + 2**12 + 2**17 + 2**19


#-------BHM SPIDERS eFEDS Clusters HSC RedMapper------ #

class BhmSpidersClusEfedsHscRedmapperCarton(BhmSpidersClusEfedsCarton):
    '''
    Carton-specific pseudo-SQL based on targetting bits in superset catalogue:
            AND WHERE x.xmatch_flags & (2^5 + 2^11) !=0
    '''

    name = 'bhm_spiders_clusters-efeds-hsc-redmapper'
    bitmask = 2**5 + 2**11


#-------BHM SPIDERS eFEDS Clusters LS RedMapper------ #

class BhmSpidersClusEfedsLsRedmapperCarton(BhmSpidersClusEfedsCarton):
    '''
    Carton-specific pseudo-SQL based on targetting bits in superset catalogue:
            AND WHERE x.xmatch_flags & (2^1 + 2^3 + 2^8 + 2^15) !=0
    '''

    name = 'bhm_spiders_clusters-efeds-ls-redmapper'
    bitmask = 2**1 + 2**3 + 2**8 + 2**15


#-------BHM SPIDERS eFEDS Clusters eROSITA------ #

class BhmSpidersClusEfedsErositaCarton(BhmSpidersClusEfedsCarton):
    '''
    Carton-specific pseudo-SQL based on targetting bits in superset catalogue:
    AND WHERE x.xmatch_flags & (2^2 + 2^6 + 2^10 + 2^12 + 2^17    +
                                2^19 + 2^5 + 2^11  +
                                2^1 + 2^3 + 2^8 + 2^15)
                                = 0
    '''

    name = 'bhm_spiders_clusters-efeds-erosita'
    bitmask =  (2**2 + 2**6 + 2**10 + 2**12 + 2**17 + 2**19 +
                2**5 + 2**11 +
                2**1 + 2**3 + 2**8 + 2**15)
    mask_sense = False





'''

 target_selection --profile tunnel_operations_sdss --verbose run --include bhm_spiders_clusters_efeds_sdss_redmapper,bhm_spiders_clusters_efeds_hsc_redmapper,bhm_spiders_clusters_efeds_ls_redmapper,bhm_spiders_clusters_efeds_erosita --keep --overwrite '0.1.0' --write-table


Exporting from the temp table

\copy (SELECT * FROM sandbox.temp_bhm_spiders_clusters_efeds_sdss_redmapper)  TO '/home/tdwelly/scratch/targetdb/bhm_spiders_clusters_efeds_sdss_redmapper.csv' with csv header
\copy (SELECT * FROM sandbox.temp_bhm_spiders_clusters_efeds_hsc_redmapper)  TO '/home/tdwelly/scratch/targetdb/bhm_spiders_clusters_efeds_hsc_redmapper.csv' with csv header
\copy (SELECT * FROM sandbox.temp_bhm_spiders_clusters_efeds_ls_redmapper)  TO '/home/tdwelly/scratch/targetdb/bhm_spiders_clusters_efeds_ls_redmapper.csv' with csv header
\copy (SELECT * FROM sandbox.temp_bhm_spiders_clusters_efeds_erosita)  TO '/home/tdwelly/scratch/targetdb/bhm_spiders_clusters_efeds_erosita.csv' with csv header


for F in bhm_spiders_clusters_*.csv; do   stilts tpipe in=${F} out="${F%.*}.fits" ifmt=csv ofmt=fits-basic; done

'''



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-06-05
# @Filename: mwm_wd.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee

from sdssdb.peewee.sdss5db.catalogdb import (Catalog, CatalogToTIC_v8,
                                             Gaia_DR2_WD, TIC_v8)

from . import BaseCarton


class MWM_WD_Carton(BaseCarton):
    """MWM White Dwarfs.

    Definition: all targets from Gentile Fusillo et al. 2019 (table
    gaia_dr2_wd) where Pwd > 0.5 and Gmag <= 20.

    """

    name = 'mwm_wd_core'
    mapper = 'MWM'
    category = 'science'
    program = 'mwm_wd'
    cadence = 'mwm_wd_2x1'
    priority = 1400

    def build_query(self, version_id, query_region=None):

        query = (Gaia_DR2_WD
                 .select(CatalogToTIC_v8.catalogid,
                         Gaia_DR2_WD.source_id,
                         Gaia_DR2_WD.pwd,
                         Gaia_DR2_WD.gmag)
                 .join(TIC_v8, on=(TIC_v8.gaia_int == Gaia_DR2_WD.source_id))
                 .join(CatalogToTIC_v8)
                 .where(Gaia_DR2_WD.pwd > self.parameters['pwd'],
                        Gaia_DR2_WD.g_gaia_mag <= self.parameters['gaia_mag'],
                        CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True))

        if query_region:
            query = (query
                     .join_from(CatalogToTIC_v8, Catalog)
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))

        return query



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-07-01
# @Filename: mwm_cb_uvex.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee
from peewee import fn

from sdssdb.peewee.sdss5db.catalogdb import Catalog, CatalogToTIC_v8, TIC_v8

from . import BaseCarton


class MWM_TESS_RGB_Carton(BaseCarton):
    """MWM TESS RGB Carton.

    Definition:

        - Jmag - Kmag > 0.5 (get red stars)
        - Hmag < 12 (get needed SNR w/APOGEE)
        - Tmag < 13 (get detections of oscillations) [Tmag := TESS magnitude]
        - M_H (absolute H band magnitude) < 1 (get giants)
            where: MH = Hmag - 10 + 5.0 * log10(parallax)
            where the parallax is in mas.

    For v0 we'll also add |b|>20

    """

    name = 'mwm_tessrgb_core'
    mapper = 'MWM'
    category = 'science'
    program = 'mwm_tessrgb'
    cadence = None
    priority = 2800

    def build_query(self, version_id, query_region=None):

        MH = TIC_v8.hmag - 10 + 5 * fn.log(TIC_v8.plx)

        query = (TIC_v8
                 .select(CatalogToTIC_v8.catalogid,
                         TIC_v8.hmag,
                         TIC_v8.jmag,
                         TIC_v8.kmag,
                         TIC_v8.tmag,
                         TIC_v8.plx)
                 .join(CatalogToTIC_v8)
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True)
                 .where((TIC_v8.jmag - TIC_v8.kmag) > 0.5)
                 .where(((TIC_v8.plx > 0) & (MH < 1)) | (TIC_v8.plx < 0),
                        fn.abs(TIC_v8.gallat) > 20)
                 .where(TIC_v8.hmag < 12)
                 .where(TIC_v8.tmag < 13))

        if query_region:
            query = (query
                     .join_from(CatalogToTIC_v8, Catalog)
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))

        return query



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-05-26
# @Filename: __main__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import importlib
import logging
import sys
import warnings

import click
import peewee

from sdssdb.peewee.sdss5db import targetdb as tdb
from sdssdb.peewee.sdss5db.catalogdb import database

import target_selection as tsmod
from target_selection.exceptions import TargetSelectionError
from target_selection.xmatch import XMatchPlanner


# Disable warnings during import because the connection may not be working yet.
with warnings.catch_warnings():
    from target_selection.cartons import BaseCarton


def all_subclasses(cls):
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in all_subclasses(c)])


def connect(profile=None, dbname=None, user=None, host=None, port=None):
    """Connects the database."""

    if profile:
        database.set_profile(profile)

    if dbname or user or host or port:
        database.connect(dbname=dbname, user=user, host=host, port=port)

    return database.connected


@click.group()
@click.option('--profile', '-p', type=str, default=None)
@click.option('--dbname', '-d', type=str, default=None)
@click.option('--user', '-u', type=str, default=None)
@click.option('--host', '-h', type=str, default=None)
@click.option('--port', '-P', type=int, default=None)
@click.option('--verbose', '-v', is_flag=True,
              help='outputs extra debug information')
@click.option('--save-log', type=str, default=None,
              help='saves the log to a file.')
def target_selection(profile, dbname, user, host, port, verbose, save_log):
    """Performs tasks related to target selection for SDSS-V."""

    if verbose:
        tsmod.log.set_level(logging.DEBUG)

    if save_log:
        tsmod.log.start_file_logger(save_log, mode='a', rotating=False)

    if not connect(profile, dbname, user, host, port):
        raise TargetSelectionError('database is not connected.')

    # Reload the cartons now that we have a connection.
    importlib.reload(sys.modules['target_selection.cartons'])


@target_selection.command()
@click.argument('TARGETING-PLAN', nargs=1, type=str)
@click.option('--config-file', type=click.Path(exists=True, dir_okay=False),
              help='the file to read. Defaults to the internal '
                   'configuration file.')
@click.option('--overwrite', is_flag=True,
              help='drop intermediate tables if they exist')
@click.option('--keep', is_flag=True,
              help='keep intermediate tables after loading')
@click.option('--skip-query', is_flag=True,
              help='do not run the query, only load intermediate '
                   'table if it exists')
@click.option('--region', '-r', type=float, nargs=3, default=None,
              help='the region (ra, dec, radius) to query')
@click.option('--load/--no-load', is_flag=True, default=True,
              help='whether to load data into targetdb.target')
@click.option('--include', '-i', type=str,
              help='comma-separated carton names to include')
@click.option('--exclude', '-e', type=str,
              help='comma-separated carton names to exclude')
@click.option('--write-table', '-w', is_flag=True,
              help='write table of loaded targets as a FITS file')
@click.option('--allow-errors', is_flag=True,
              help='continue processing cartons if a carton fails')
def run(targeting_plan, config_file, overwrite, keep, region, load,
        skip_query, include, exclude, write_table, allow_errors):
    """Runs target selection for all cartons."""

#    carton_classes = {Carton.name: Carton
#                      for Carton in BaseCarton.__subclasses__()}
    # find grandchilden Cartons as well
    carton_classes = {Carton.name: Carton
                      for Carton in all_subclasses(BaseCarton)}

    if len(carton_classes) == 0:
        raise TargetSelectionError('no carton classes found.')

    try:
        config_plan = tsmod.config[targeting_plan]
    except KeyError:
        raise TargetSelectionError('cannot find configuration for plan '
                                   f'{targeting_plan}.')

    cartons = config_plan['cartons']

    if exclude:
        cartons = [carton for carton in cartons if carton not in exclude]

    if include:
        cartons = [carton for carton in cartons if carton in include]

    for carton_name in cartons:

        try:

            Carton = carton_classes[carton_name]
            carton = Carton(targeting_plan, config_file=config_file)

            tsmod.log.header = f'({carton.name}): '
            tsmod.log.info(f'Running target selection for '
                           f'carton {carton.name!r}.')

            if carton.check_targets() and not overwrite:
                raise ValueError(f'Found existing targets for carton '
                                 f'{carton.name!r} with plan {carton.plan!r}.')

            if not skip_query:
                carton.run(query_region=(region or None), overwrite=overwrite)
                if write_table:
                    carton.write_table()
            else:
                tsmod.log.debug('Skipping query.')

            if load:
                carton.load(overwrite=overwrite)
            else:
                tsmod.log.info('Not loading data into targetdb.target.')

            if write_table:
                carton.write_table(mode='targetdb')

            if not keep:
                tsmod.log.info(f'Dropping temporary table {carton.path!r}.')
                carton.drop_table()

        except Exception as ee:

            if allow_errors:
                if carton_name not in carton_classes:
                    tsmod.log.error(f'No carton class found for {carton_name}')
                else:
                    tsmod.log.error(f'Errored processing carton '
                                    f'{carton_name}: {ee}')
            else:
                raise

    tsmod.log.header = ''


@target_selection.command()
@click.argument('TARGETING-PLAN', type=str)
@click.option('--tables', is_flag=True,
              help='also remove intermediate tables')
def clear(targeting_plan, tables):
    """Clear all data for a target selection plan."""

    if tables:
        # Cartons = BaseCarton.__subclasses__()
        # find grandchildren classes as well
        Cartons = all_subclasses(BaseCarton)
        for Carton in Cartons:
            Carton(targeting_plan).drop_table()
    try:
        tdb.Version.get(plan=targeting_plan,
                        target_selection=True).delete_instance()
    except peewee.DoesNotExist:
        pass


@target_selection.command()
@click.argument('XMATCH-PLAN', type=str)
@click.option('--file', type=click.Path(exists=True, dir_okay=False),
              help='the file to read. Defaults to the internal '
                   'configuration file.')
def xmatch(xmatch_plan, file):
    """Runs catalogue cross-matching from a configuration file."""

    xmatch = XMatchPlanner.read(database, xmatch_plan, config_file=file)
    xmatch.run()


if __name__ == '__main__':

    target_selection()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Tom Dwelly
# @Date: 2020-03-03
# @Filename: bhm_aqmes.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

# flake8: noqa
# isort: skip_file

import peewee
import sdssdb
from astropy.io import fits
import pkg_resources

#debug
#import os


from sdssdb.peewee.sdss5db.catalogdb import (Catalog,
                                             SDSS_DR16_SpecObj,
                                             SDSS_DR16_QSO,
                                             CatalogToSDSS_DR16_SpecObj)


from target_selection.cartons.base import BaseCarton

# this should be in a better place
radius_apo = 1.49 # degrees


# Details: Start here
# https://wiki.sdss.org/display/OPS/Defining+target+selection+and+cadence+algorithms#Definingtargetselectionandcadencealgorithms-AQMES-medium-inprogress

## This provides the following BHM cartons:
# bhm_aqmes_med
# bhm_aqmes_med-faint
# bhm_aqmes_wide2
# bhm_aqmes_wide2-faint
# bhm_aqmes_wide3
# bhm_aqmes_wide3-faint
# bhm_aqmes_bonus-dark
# bhm_aqmes_bonus-bright



class BhmAqmesBaseCarton(BaseCarton):
    ''' Parent class that provides the underlying selections for all AQMES cartons'''

    name = 'bhm_aqmes_base'
    category = 'science'
    mapper = 'BHM'
    program = 'bhm_aqmes'
    tile = False
    priority = None
    alias_c = None
    alias_t = None
    alias_c2s = None

    # read the AQMES field centres from a fits file and convert to a list of dicts
    def get_fieldlist(self):
        stub = self.parameters.get('fieldlist', None)
        if stub is None or stub == '' or stub == 'None':
            return None

        #filename = pkg_resources.resource_filename( __name__, stub)
        filename = pkg_resources.resource_filename('target_selection', stub)
        assert len(filename) > 0

        try:
            hdul = fits.open(filename)
        except:
            raise Exception(f"Failed to find/open fieldlist file: {filename}")

        assert len(hdul[1].data) > 0

        # choose the correct subset of fields based on the cadence name and form a list of dicts
        try:
            fieldlist = [ {'racen': r['RACEN'],
                           'deccen': r['DECCEN'],
                           'radius': radius_apo, }
                          for r in hdul[1].data
                          if r['CADENCE'] == self.cadence
            ]
        except:
            raise Exception(f"Error interpreting contents of fieldlist file: {filename}")

        assert len(fieldlist) > 0

        return fieldlist

    def append_spatial_query(self, query, fieldlist):
        '''extend the peewee query using a list of field centres'''
        if fieldlist is None :
            return query
        elif len(fieldlist) == 0 :
            return query

        q = False
        for f in fieldlist:
            q = ( q | peewee.fn.q3c_radial_query(self.alias_c.ra,
                                                 self.alias_c.dec,
                                                 f['racen'],
                                                 f['deccen'],
                                                 f['radius']))
        return query.where(q)


    def build_query(self, version_id, query_region=None):
        c = Catalog.alias()
        c2s = CatalogToSDSS_DR16_SpecObj.alias()
        s = SDSS_DR16_SpecObj.alias()
        t = SDSS_DR16_QSO.alias()
        self.alias_c = c
        self.alias_t = t
        self.alias_c2s = c2s

        # set the Carton priority+values here - read from yaml
        target_priority = peewee.Value(int(self.parameters.get('priority', 10000))).alias('priority')
        target_value = peewee.Value(self.parameters.get('value', 1.0)).cast('float').alias('value')
        pmra =  peewee.Value(0.0).cast('float').alias('pmra')
        pmdec = peewee.Value(0.0).cast('float').alias('pmdec')
        parallax = peewee.Value(0.0).cast('float').alias('parallax')

        query = (
            c
            .select(c.catalogid,
                    target_priority,
                    target_value,
                    pmra,
                    pmdec,
                    parallax,
                    t.psfmag[1].alias('g'),
                    t.psfmag[2].alias('r'),
                    t.psfmag[3].alias('i'),
                    t.psfmag[4].alias('z'),
            )
            .join(c2s)
            .join(s)
            .join(t, on=((s.plate == t.plate) &
                         (s.mjd == t.mjd) &
                         (s.fiberid == t.fiberid)))
            .where(c.version_id == version_id,
                   c2s.version_id == version_id)
#                   c2s.best == True)  # this was removing many aqmes-med targets
            .distinct([t.pk])   # avoid duplicates - trust the QSO parent sample
            .where
            (
                (t.psfmag[3] >= self.parameters['mag_i_min']),
                (t.psfmag[3] <  self.parameters['mag_i_max']),
                #            (t.z >= self.parameters['redshift_min']), # not needed
                #            (t.z <= self.parameters['redshift_max']),
            )
        )
        query = self.append_spatial_query(query, self.get_fieldlist())


        return query



#-------AQMES medium ------ #

class BhmAqmesMedCarton(BhmAqmesBaseCarton):
    '''
    SELECT * FROM sdss_dr1x_qso WHERE  psfmag_i BETWEEN 16.x AND 19.1
    '''
    name = 'bhm_aqmes_med'
    cadence = 'bhm_aqmes_medium_12x4'
    #cadence = 'dummy_cadence'

# add something like the following if want to add carton-specific selections
#    def build_query(self, version_id, query_region=None):
#        query = super().build_query(version_id, query_region)
#        query = query.where( # .... add extra terms here
#        )
#        return query


class BhmAqmesMedFaintCarton(BhmAqmesBaseCarton):
    '''
    SELECT * FROM sdss_dr1x_qso WHERE  psfmag_i BETWEEN 19.1 AND 21.0
    '''
    name = 'bhm_aqmes_med-faint'
    cadence = 'bhm_aqmes_medium_12x4'
    program = 'bhm_filler'

#-------AQMES medium ------ #


#-------AQMES wide ------ #

class BhmAqmesWide3Carton(BhmAqmesBaseCarton):
    '''
    SELECT * FROM sdss_dr1x_qso WHERE psfmag_i BETWEEN 16.x AND 19.1
    '''
    name = 'bhm_aqmes_wide3'
    cadence = 'bhm_aqmes_wide_3x4'


class BhmAqmesWide3FaintCarton(BhmAqmesBaseCarton):
    '''
    SELECT * FROM sdss_dr1x_qso WHERE psfmag_i BETWEEN 19.1 AND 21.0
    '''
    name = 'bhm_aqmes_wide3-faint'
    cadence = 'bhm_aqmes_wide_3x4'
    program = 'bhm_filler'


class BhmAqmesWide2Carton(BhmAqmesBaseCarton):
    '''
    SELECT * FROM sdss_dr1x_qso WHERE psfmag_i BETWEEN 16.x AND 19.1
    '''
    name = 'bhm_aqmes_wide2'
    cadence = 'bhm_aqmes_wide_2x4'


class BhmAqmesWide2FaintCarton(BhmAqmesBaseCarton):
    '''
    SELECT * FROM sdss_dr1x_qso WHERE psfmag_i BETWEEN 19.1 AND 21.0
    '''
    name = 'bhm_aqmes_wide2-faint'
    cadence = 'bhm_aqmes_wide_2x4'
    program = 'bhm_filler'

#-------AQMES wide ------ #


#-------AQMES bonus ------ #

class BhmAqmesBonusDarkCarton(BhmAqmesBaseCarton):
    '''
    SELECT * FROM sdss_dr1x_qso WHERE psfmag_i BETWEEN 16.x AND 21.5
    '''
    name = 'bhm_aqmes_bonus-dark'
    cadence = 'bhm_spiders_1x4'
    program = 'bhm_filler'

    # add carton-specific selections - this prevents a problem with duplicated cross-matches
    # applying this down-selection to other AQMES cartons reduces the numbers of targets significantly.
    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        query = query.where(self.alias_c2s.best == True)
        return query


class BhmAqmesBonusBrightCarton(BhmAqmesBaseCarton):
    '''
    SELECT * FROM sdss_dr1x_qso WHERE psfmag_i BETWEEN 14.0 AND 18.0
    '''
    name = 'bhm_aqmes_bonus-bright'
    cadence = 'bhm_boss_bright_3x1'   # could add a new cadence name for this
    program = 'bhm_filler'

#-------AQMES bonus ------ #








##################################################################################
### Debug stuff

'''
SQL:
select c.*,s.specobjid,t.class,t.psfmag[2] as psfmag_g,t.psfmag[4] AS psfmag_i,t.z,t.zwarning,t.plate,t.mjd,t.fiberid,q3c_dist(c.ra,c.dec,t.ra,t.dec)*3600.::FLOAT as sep from catalog AS C INNER JOIN catalog_to_sdss_dr13_photoobj as c2p ON c.catalogid = c2p.catalogid INNER JOIN sdss_dr13_photoobj AS p ON c2p.target_id = p.objid INNER JOIN sdss_dr16_specobj as s on p.objid = s.bestobjid INNER JOIN sdss_dr14_qso AS t ON (s.plate = t.plate AND s.mjd = t.mjd AND s.fiberid = t.fiberid ) WHERE (q3c_radial_query(c.ra, c.dec, 10.0, 25.0, 0.1) AND c.version_id = 13 AND (p.resolvestatus & 256) != 0 AND t.psfmag[4] < 19.1);

#python:
t = SDSS_DR14_QSO.alias()
for f in t._meta.fields:
     print (f)

'''

'''
# for testing do domething like this

import peewee
import sdssdb
from sdssdb.peewee.sdss5db.catalogdb import database
database.set_profile('tunnel_operations')
from target_selection.cartons.bhm_aqmes import *
c = BhmAqmesMedCarton(targeting_plan='0.1.0-beta.1')
q = c.build_query(version_id=13)
for r in q.limit(5).namedtuples():
    print(r)

'''


'''
target_selection --profile tunnel_operations_sdss --verbose run --include bhm_aqmes_med,bhm_aqmes_med_faint,bhm_aqmes_wide2,bhm_aqmes_wide3,bhm_aqmes_wide2_faint,bhm_aqmes_wide3_faint,bhm_aqmes_bonus_dark,bhm_aqmes_bonus_bright --keep --overwrite '0.1.0' --write-table

# Exporting from the temp table

\copy (SELECT * FROM sandbox.temp_bhm_aqmes_med)  TO '/home/tdwelly/scratch/targetdb/bhm_aqmes_med.csv' with csv header
\copy (SELECT * FROM sandbox.temp_bhm_aqmes_med_faint)  TO '/home/tdwelly/scratch/targetdb/bhm_aqmes_med_faint.csv' with csv header

\copy (SELECT * FROM sandbox.temp_bhm_aqmes_wide2)  TO '/home/tdwelly/scratch/targetdb/bhm_aqmes_wide2.csv' with csv header
\copy (SELECT * FROM sandbox.temp_bhm_aqmes_wide3)  TO '/home/tdwelly/scratch/targetdb/bhm_aqmes_wide3.csv' with csv header
\copy (SELECT * FROM sandbox.temp_bhm_aqmes_wide2_faint)  TO '/home/tdwelly/scratch/targetdb/bhm_aqmes_wide2_faint.csv' with csv header
\copy (SELECT * FROM sandbox.temp_bhm_aqmes_wide3_faint)  TO '/home/tdwelly/scratch/targetdb/bhm_aqmes_wide3_faint.csv' with csv header
\copy (SELECT * FROM sandbox.temp_bhm_aqmes_bonus_bright)  TO '/home/tdwelly/scratch/targetdb/bhm_aqmes_bonus_bright.csv' with csv header
\copy (SELECT * FROM sandbox.temp_bhm_aqmes_bonus_dark)  TO '/home/tdwelly/scratch/targetdb/bhm_aqmes_bonus_dark.csv' with csv header

for F in bhm_aqmes_*.csv; do   stilts tpipe in=${F} out="${F%.*}.fits" ifmt=csv ofmt=fits-basic; done

'''

###################################################################################



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-02-24
# @Filename: base.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import abc
import inspect
import warnings

import numpy
import peewee
from astropy import table

from sdssdb.peewee import BaseModel
from sdssdb.peewee.sdss5db import catalogdb as cdb
from sdssdb.peewee.sdss5db import targetdb as tdb
from sdsstools import read_yaml_file
from sdsstools.color_print import color_text

from target_selection import __version__, config, log
from target_selection.exceptions import (TargetSelectionError,
                                         TargetSelectionUserWarning)
from target_selection.utils import Timer


EPOCH = 2015.5


class BaseCarton(metaclass=abc.ABCMeta):
    """A base class for target cartons.

    This class is not intended for direct instantiation. Instead, it must be
    subclassed and the relevant class attributes overridden with the values
    corresponding to the carton.

    Parameters
    ----------
    targeting_plan : str
        The target selection plan version.
    config_file : str
        The path to the configuration file to use. If undefined, uses the
        internal ``target_selection.yml`` file.
    schema : str
        Schema in which the temporary table with the results of the
        query will be created. If `None`, tries to use the ``schema`` parameter
        from the configuration file for this plan of target selection. If
        the parameter is not set, defaults to ``'sandbox'``.
    table_name : str
        The name of the temporary table. Defaults to ``temp_<name>`` where
        ``<name>`` is the target class name.

    Attributes
    ----------
    name : str
        The name of the carton (required).
    cadence : str
        The label of the cadence rule for this carton.
    category : str
        The category of targets for this carton.
    mapper : str
        The mapper with which this carton is associated.
    orm : str
        The ORM library to be used, ``peewee`` or ``sqlalchemy``.
    tag : str
        The version of the ``target_selection`` code used.
    query_region : tuple
        A tuple defining the region over which the query should be performed,
        with the format ``(ra, dec, radius)`` in degrees. This will append a
        ``q3c_radial_query`` condition to the query.
    load_magnitudes : bool
        Whether to load target magnitudes. In general this must be `True`
        except for cartons for which it's known the magnitudes will not be
        used, e.g., skies.

    """

    name = None
    cadence = None
    category = None
    program = None
    mapper = None
    priority = None

    load_magnitudes = True

    query_region = None

    def __init__(self, targeting_plan, config_file=None,
                 schema=None, table_name=None):

        assert self.name, 'carton subclass must override name'
        assert self.category, 'carton subclass must override category'
        assert self.program, 'carton subclass must override program'

        self.plan = targeting_plan
        self.tag = __version__

        if config_file:
            this_config = read_yaml_file(config_file)
        else:
            this_config = config

        if self.plan not in this_config:
            raise TargetSelectionError(f'({self.name}): cannot find plan '
                                       f'{self.plan!r} in config.')

        self.config = this_config[self.plan]

        if 'parameters' in self.config:
            self.parameters = self.config['parameters'].get(self.name, None)
        else:
            self.parameters = None

        try:
            self.xmatch_plan = self.config['xmatch_plan']
        except KeyError:
            raise TargetSelectionError(f'({self.name}): xmatch_plan '
                                       'not found in config.')

        # Check the signature of build_query
        self._build_query_signature = inspect.signature(self.build_query)
        if 'version_id' not in self._build_query_signature.parameters:
            raise TargetSelectionError('build_query does not '
                                       'accept version_id')

        self.database = tdb.database
        assert self.database.connected, 'database is not connected.'

        self.schema = schema or self.config.get('schema', None) or 'sandbox'
        self.table_name = table_name or f'temp_{self.name}'

        if self.cadence:
            assert (tdb.Cadence
                    .select()
                    .where(tdb.Cadence.label == self.cadence)
                    .count() == 1), (f'{self.cadence!r} does not '
                                     'exist in targetdb.cadence.')

        self.log = log
        self.has_run = False

    @property
    def path(self):
        """The schema-qualified path to the output table."""

        return self.schema + '.' + self.table_name

    @abc.abstractmethod
    def build_query(self, version_id, query_region=None):
        """Builds and returns the query.

        The ORM query for the target class. Note that this must be the
        *un-executed* query, which will be executed in `.run`. The select
        statement must be run on ``catalogdb.catalog`` and needs to include,
        at least, ``catalogid``. Additional columns such as
        ``ra, dec, pmra, pmdec`` can be specified but are otherwise completed
        when calling `.run`. Magnitude columns can be included and propagated
        to ``targetdb`` but must be aliased as ``magnitude_<band>`` where
        ``<band>`` must be one of the columns in ``targetdb.magnitude``.

        The query returned must include a filter for ``version_id`` on
        ``catalogdb.catalog``. Normally this is applied to the ``Catalog``
        select as ``.where(Catalog.version_id == version_id)``.

        Parameters
        ----------
        version_id : int
            The id of the cross-match version to use.

        Returns
        -------
        query
            A :class:`peewee:Select` or :class:`peewee:ModelSelect` query.

        """

        pass

    def get_model(self):
        """Returns a Peewee model for the temporary table using reflection."""

        class Model(BaseModel):

            catalogid = peewee.BigIntegerField(primary_key=True)
            selected = peewee.BooleanField()
            cadence = peewee.TextField(null=True)
            priority = peewee.IntegerField()

            class Meta:
                database = self.database
                table_name = self.table_name
                schema = self.schema
                reflection_options = {'skip_foreign_keys': True}
                use_reflection = True

        if not Model.table_exists():
            raise TargetSelectionError(f'temporary table {self.path!r} does '
                                       'not exist.')

        return Model

    def run(self, tile=None, query_region=None, overwrite=False,
            **post_process_kawrgs):
        """Executes the query and post-process steps, and stores the results.

        This method calls `.build_query` and runs the returned query. The
        output of the query is stored in a temporary table whose schema and
        table name are defined when the object is instantiated.

        After the query has run, the `.post_process` routine is called if the
        method has been overridden for the given carton.

        Parameters
        ----------
        query_region : tuple
            A tuple defining the region over which the query should be
            performed, with the format ``(ra, dec, radius)`` in degrees. This
            will append a ``q3c_radial_query`` condition to the query.
        overwrite : bool
            Whether to overwrite the intermediary table if already exists.
        post_process_args : dict
            Keyword arguments to be passed to `.post_process`.

        Returns
        -------
        model : :class:`peewee:Model`
            The model for the intermediate table.

        """

        query_region = query_region or self.query_region

        if self.database.table_exists(self.table_name, schema=self.schema):
            if overwrite:
                log.info(f'Dropping table {self.path!r}.')
                self.drop_table()
            else:
                raise RuntimeError(f'Temporary table {self.path!r} '
                                   'already exists.')

        log.info('Running query ...')
        version_id = cdb.Version.get(plan=self.xmatch_plan).id

        # If build_query accepts a query_region parameter, call with the query
        # region. Otherwise will add the radial query condition later.
        if 'query_region' in self._build_query_signature.parameters:
            query = self.build_query(version_id, query_region=query_region)
        else:
            query = self.build_query(version_id)

        # Make sure the catalogid column is selected.
        if cdb.Catalog.catalogid not in query._returning:
            raise RuntimeError('catalogid is not being returned in query.')

        if query_region:
            if 'query_region' in self._build_query_signature.parameters:
                pass
            else:
                # This may be quite inefficient depending on the query.
                subq = query.alias('subq')
                query = (peewee.Select(columns=[peewee.SQL('subq.*')])
                         .from_(subq)
                         .join(cdb.Catalog,
                               on=(cdb.Catalog.catalogid == subq.c.catalogid))
                         .where(peewee.fn.q3c_radial_query(cdb.Catalog.ra,
                                                           cdb.Catalog.dec,
                                                           query_region[0],
                                                           query_region[1],
                                                           query_region[2])))

        query_sql, params = query.sql()
        cursor = self.database.cursor()
        query_str = cursor.mogrify(query_sql, params).decode()

        log.debug(color_text(f'CREATE TABLE IF NOT EXISTS {self.path} AS ' +
                             query_str, 'darkgrey'))

        with self.database.atomic():
            with Timer() as timer:
                self.setup_transaction()
                self.database.execute_sql(f'CREATE TABLE IF NOT EXISTS '
                                          f'{self.path} AS ' + query_sql,
                                          params)

        log.info(f'Created table {self.path!r} in {timer.interval:.3f} s.')

        log.debug('Adding columns and indexes.')

        columns = [col.name
                   for col in self.database.get_columns(self.table_name,
                                                        self.schema)]

        if 'selected' not in columns:
            self.database.execute_sql(f'ALTER TABLE {self.path} '
                                      'ADD COLUMN selected BOOL DEFAULT TRUE;')

        if 'cadence' not in columns:
            self.database.execute_sql(f'ALTER TABLE {self.path} '
                                      'ADD COLUMN cadence VARCHAR DEFAULT NULL;')
        if 'priority' not in columns:
            self.database.execute_sql(f'ALTER TABLE {self.path} '
                                      'ADD COLUMN priority INTEGER '
                                      'DEFAULT NULL;')

        self.database.execute_sql(f'ALTER TABLE {self.path} '
                                  'ADD PRIMARY KEY (catalogid);')
        self.database.execute_sql(f'CREATE INDEX ON {self.path} (selected);')
        self.database.execute_sql(f'ANALYZE {self.path};')

        ResultsModel = self.get_model()

        n_rows = ResultsModel.select().count()
        log.debug(f'Table {self.path!r} contains {n_rows:,} rows.')

        log.debug('Running post-process.')
        with self.database.atomic():
            self.setup_transaction()
            self.post_process(ResultsModel, **post_process_kawrgs)

        self.has_run = True

        return ResultsModel

    def post_process(self, model, **kwargs):
        """Post-processes the temporary table.

        This method provides a framework for applying non-SQL operations on
        carton query. It receives the model for the temporary table and can
        perform any operation on it, including modifying the ``selected``
        column with a mask of targets to be used.

        This method can also be used to set the ``cadence`` column in the
        temporary table. This column will be used to set the target cadence if
        the carton `.cadence` attribute is not set.

        `.post_process` runs inside a database transaction so it's not
        necessary to create a new one, but savepoints can be added.

        Parameters
        ----------
        model : peewee:Model
            The model of the intermediate table.

        Returns
        -------
        mask : `tuple`
            The list of catalogids from the temporary table that should be
            selected as part of this carton. If `True` (the default), selects
            all the records.

        """

        return True

    def setup_transaction(self):
        """Setups the transaction locally modifying the datbase parameters.

        This method runs inside a transaction and can be overridden to set the
        parameters of the transaction manually. It applies to both `.run` and
        `.load`.

        """

        if 'database_options' not in self.config:
            return

        for param, value in self.config['database_options'].items():
            self.database.execute_sql(f'SET LOCAL {param} = {value!r};')

    def drop_table(self):
        """Drops the intermediate table if it exists."""

        self.database.execute_sql(f'DROP TABLE IF EXISTS {self.path};')

    def write_table(self, filename=None, mode='results', write=True):
        """Writes the selection to a FITS file.

        Parameters
        ----------
        filename : str
            The file to which to write the table. Defaults to
            ``<name>_<plan>.fits``.
        mode : str
            Defines what data to write. If ``'results'``, writes the
            intermediate table (usually just the ``catalogid`` column). If
            ``'targetdb'``, writes all the relevant columns for the targets
            loaded to ``targetdb`` for this carton and plan (must be used after
            `.load` has been called).
        write : bool
            Whether to write the table to disk. If `False`, just returns the
            table object.

        Returns
        -------
        table : `~astropy.table.Table`
            A table object with the selected results.

        """

        if not filename:
            if mode == 'results':
                filename = f'{self.name}_{self.plan}.fits.gz'
            else:
                filename = f'{self.name}_{self.plan}_targetdb.fits.gz'

        log.debug(f'Writing table to {filename}.')

        if mode == 'results':

            results_model = self.get_model()
            assert results_model.table_exists(), \
                'temporary table does not exist.'

            write_query = results_model.select()

            colnames = [field.name for field in write_query._returning]

        elif mode == 'targetdb':

            mag_fields = [field
                          for field in tdb.Magnitude._meta.fields.values()
                          if field.name not in ['pk', 'target_pk', 'target']]

            write_query = (tdb.Target
                           .select(tdb.Target,
                                   *mag_fields,
                                   tdb.CartonToTarget.priority,
                                   tdb.Cadence.label.alias('cadence'))
                           .join(tdb.CartonToTarget)
                           .join(tdb.Magnitude)
                           .join_from(tdb.CartonToTarget, tdb.Cadence,
                                      peewee.JOIN.LEFT_OUTER)
                           .join_from(tdb.CartonToTarget, tdb.Carton)
                           .join(tdb.Version)
                           .where(tdb.Carton.carton == self.name,
                                  tdb.Version.plan == self.plan,
                                  tdb.Version.target_selection >> True))

            colnames = []
            for col in write_query._returning:
                if isinstance(col, peewee.ForeignKeyField):
                    colnames.append(col.column_name)
                elif isinstance(col, peewee.Alias):
                    colnames.append(col._alias)
                else:
                    colnames.append(col.name)

        else:
            raise ValueError('invalud mode. Available modes are '
                             '"results" and "targetdb".')

        if not write_query.exists():
            raise TargetSelectionError('no records found.')

        results = write_query.tuples()
        results = ((col if col is not None else numpy.nan for col in row)
                   for row in tuple(results))

        warnings.filterwarnings(
            'ignore', message='.*converting a masked element to nan.*')

        carton_table = table.Table(rows=results, names=colnames, masked=True)

        if write:
            carton_table.write(filename, overwrite=True)

        return carton_table

    def load(self, overwrite=False):
        """Loads the output of the intermediate table into targetdb."""

        if self.check_targets():
            if overwrite:
                warnings.warn(f'Carton {self.name!r} with plan {self.plan!r} '
                              f'and tag {self.tag!r} already has targets '
                              'loaded. Dropping carton-to-target entries.',
                              TargetSelectionUserWarning)
                self.drop_carton()
            else:
                raise TargetSelectionError(f'Found existing targets for '
                                           f'carton {self.name!r} with plan '
                                           f'{self.plan!r} and tag '
                                           f'{self.tag!r}.')

        RModel = self.get_model()
        if not RModel.table_exists():
            raise TargetSelectionError(f'No temporary table found '
                                       f'{self.full}. Did you call run()?')

        has_targets = (RModel.select()
                       .where(RModel.selected >> True)
                       .exists())

        if not has_targets:
            raise TargetSelectionError('No targets found in '
                                       'intermediate table.')

        with self.database.atomic():
            self.setup_transaction()
            self._create_carton_metadata()
            self._load_targets(RModel)
            self._load_carton_to_target(RModel)
            if self.load_magnitudes:
                self._load_magnitudes(RModel)
            else:
                warnings.warn('Skipping magnitude load.',
                              TargetSelectionUserWarning)

            self.log.debug('Committing records and checking constraints.')

    def check_targets(self):
        """Check if data has been loaded for this carton and targeting plan."""

        has_targets = (tdb.CartonToTarget
                       .select()
                       .join(tdb.Carton)
                       .join(tdb.Version)
                       .where(tdb.Carton.carton == self.name,
                              tdb.Version.plan == self.plan,
                              tdb.Version.target_selection >> True)
                       .exists())

        return has_targets

    def _create_carton_metadata(self):

        mapper_pk = None
        category_pk = None

        # Create targeting plan in tdb.
        version, created = tdb.Version.get_or_create(plan=self.plan,
                                                     tag=self.tag,
                                                     target_selection=True)
        version_pk = version.pk

        if created:
            log.info(f'Created record in targetdb.version for '
                     f'{self.plan!r} with tag {self.tag!r}.')

        if (tdb.Carton.select()
                      .where(tdb.Carton.carton == self.name,
                             tdb.Carton.version_pk == version_pk)
                      .exists()):
            return

        # Create carton and associated values.
        if self.mapper:
            mapper, created_pk = tdb.Mapper.get_or_create(label=self.mapper)
            mapper_pk = mapper.pk
            if created:
                log.debug(f'Created mapper {self.mapper!r}')

        if self.category:
            category, created = tdb.Category.get_or_create(label=self.category)
            category_pk = category.pk
            if created:
                log.debug(f'Created category {self.category!r}')

        tdb.Carton.create(carton=self.name, category_pk=category_pk,
                          program=self.program, mapper_pk=mapper_pk,
                          version_pk=version_pk).save()

        log.debug(f'Created carton {self.name!r}')

    def _load_targets(self, RModel):
        """Load data from the intermediate table tp targetdb.target."""

        log.debug('loading data into targetdb.target.')

        n_inserted = tdb.Target.insert_from(
            cdb.Catalog.select(cdb.Catalog.catalogid,
                               cdb.Catalog.ra,
                               cdb.Catalog.dec,
                               cdb.Catalog.pmra,
                               cdb.Catalog.pmdec,
                               cdb.Catalog.parallax,
                               peewee.Value(EPOCH))
            .join(RModel, on=(cdb.Catalog.catalogid == RModel.catalogid))
            .where(RModel.selected >> True)
            .where(~peewee.fn.EXISTS(
                tdb.Target
                .select(peewee.SQL('1'))
                .where(tdb.Target.catalogid == RModel.catalogid))),
            [tdb.Target.catalogid,
             tdb.Target.ra,
             tdb.Target.dec,
             tdb.Target.pmra,
             tdb.Target.pmdec,
             tdb.Target.parallax,
             tdb.Target.epoch]).returning().execute()

        log.info(f'Inserted {n_inserted:,} new rows into targetdb.target.')

        return

    def _load_magnitudes(self, RModel):
        """Load magnitudes into targetdb.magnitude."""

        log.debug('Loading data into targetdb.magnitude.')

        Magnitude = tdb.Magnitude

        magnitude_paths = self.config['magnitudes']
        fields = [Magnitude.carton_to_target_pk]

        select_from = (RModel
                       .select(tdb.CartonToTarget.pk)
                       .join(tdb.Target,
                             on=(RModel.catalogid == tdb.Target.catalogid))
                       .join(tdb.CartonToTarget)
                       .join(tdb.Carton)
                       .join(tdb.Version)
                       .where(RModel.selected >> True)
                       .where(tdb.Carton.carton == self.name,
                              tdb.Version.plan == self.plan,
                              tdb.Version.tag == self.tag,
                              tdb.Version.target_selection >> True))

        for mag, mpath in magnitude_paths.items():

            fields.append(getattr(Magnitude, mag))
            if hasattr(RModel, mag):
                select_from = select_from.select_extend(getattr(RModel, mag))
                continue

            select_from = select_from.switch(tdb.Target)

            # For each node in the join list we check if the node model has
            # already been join and if so, switch the pointer to that model.
            # Otherwise do a LEFT OUTER join because we want all the rows in
            # the temporary table even if they don't have associated
            # magnitudes. We make sure we only use the "best" match from the
            # cross-match. No need to apply filters on cross-match version_id
            # because catalogid is unique across versions.

            for node in mpath:
                column = None
                if node == mpath[-1]:
                    node, column = node.split('.')
                node_model = self.database.models.get('catalogdb.' + node)
                joins = [model[0] for join in select_from._joins.values()
                         for model in join]
                if node_model in joins:
                    select_from = select_from.switch(node_model)
                else:
                    if node.startswith('catalog_to_'):
                        select_from = select_from.join(
                            node_model,
                            peewee.JOIN.LEFT_OUTER,
                            on=(tdb.Target.catalogid == node_model.catalogid))
                        select_from = (select_from
                                       .where((node_model.best >> True) |
                                              (node_model.catalogid >> None)))
                    else:
                        select_from = select_from.join(node_model,
                                                       peewee.JOIN.LEFT_OUTER)
                if column:
                    select_from = (select_from
                                   .select_extend(getattr(node_model, column)))

        n_inserted = (Magnitude
                      .insert_from(select_from, fields)
                      .returning().execute())

        log.info(f'Inserted {n_inserted:,} new rows into targetdb.magnitude.')

    def _load_carton_to_target(self, RModel):
        """Populate targetdb.carton_to_target."""

        log.debug('Loading data into targetdb.carton_to_target.')

        version_pk = tdb.Version.get(plan=self.plan, tag=self.tag,
                                     target_selection=True)
        carton_pk = tdb.Carton.get(carton=self.name, version_pk=version_pk).pk

        Target = tdb.Target
        CartonToTarget = tdb.CartonToTarget

        select_from = (RModel
                       .select(Target.pk,
                               carton_pk)
                       .join(Target,
                             on=(Target.catalogid == RModel.catalogid))
                       .where(RModel.selected >> True)
                       .where(~peewee.fn.EXISTS(
                           CartonToTarget
                           .select(peewee.SQL('1'))
                           .join(tdb.Carton)
                           .where(CartonToTarget.target_pk == Target.pk,
                                  CartonToTarget.carton_pk == carton_pk,
                                  tdb.Carton.version_pk == version_pk))))

        if self.cadence is not None:

            # Check that not both the carton cadence and the cadence column
            # are not null.
            if RModel.select().where(~(RModel.cadence >> None)).exists():
                raise TargetSelectionError('both carton cadence and target '
                                           'cadence defined. This is not '
                                           'allowed.')

            cadence_pk = tdb.Cadence.get(label=self.cadence)
            select_from = select_from.select_extend(cadence_pk)

        else:

            # If all cadences are null we'll set that as a value and save us
            # a costly join.
            if not RModel.select().where(~(RModel.cadence >> None)).exists():
                select_from = select_from.select_extend(peewee.SQL('null'))
            else:
                select_from = (select_from
                               .select_extend(tdb.Cadence.pk)
                               .switch(RModel)
                               .join(tdb.Cadence, 'LEFT OUTER JOIN',
                                     on=(tdb.Cadence.label == RModel.cadence)))

        if self.priority is None:
            select_from = select_from.select_extend(RModel.priority)
        else:
            select_from = select_from.select_extend(self.priority)

        # Now do the insert
        n_inserted = CartonToTarget.insert_from(
            select_from,
            [CartonToTarget.target_pk,
             CartonToTarget.carton_pk,
             CartonToTarget.cadence_pk,
             CartonToTarget.priority]).returning().execute()

        log.info(f'Inserted {n_inserted:,} rows '
                 'into targetdb.carton_to_target.')

    def drop_carton(self):
        """Drops the entry in ``targetdb.carton``."""

        version = (tdb.Version
                   .select()
                   .where(tdb.Version.plan == self.plan,
                          tdb.Version.tag == self.tag,
                          tdb.Version.target_selection >> True))

        if version.count() == 0:
            return

        tdb.Carton.delete().where(tdb.Carton.carton == self.name,
                                  tdb.Carton.version == version).execute()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-06-16
# @Filename: mwm_dust.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import numpy
import pandas
import peewee

from sdssdb.peewee.sdss5db import targetdb
from sdssdb.peewee.sdss5db.catalogdb import (GLIMPSE, AllWise, Catalog,
                                             CatalogToAllWise,
                                             CatalogToGLIMPSE, CatalogToTIC_v8,
                                             TIC_v8, TwoMassPSC)

from . import BaseCarton


def lbp2xyz(ll, bb, pp):

    dist = 1. / pp  # kpc
    z = dist * numpy.sin(numpy.radians(bb))
    x = (dist * numpy.cos(numpy.radians(bb)) * numpy.cos(numpy.radians(ll)))
    y = (dist * numpy.cos(numpy.radians(bb)) * numpy.sin(numpy.radians(ll)))

    # z is definitely right; x and y depend on conventions.  As long
    # as this routine is used for x and y throughout this code, this should
    # not introduce any issues.

    return x, y, z


def subselect(data, othersel, downsampledby=1):
    """Turn x, y, z into pixel; bin up to see where we are missing."""

    ll = data['gallong']
    bb = data['gallat']
    xyz = lbp2xyz(ll, bb, data['plx'])

    dustsel = numpy.ones(len(data), numpy.bool)

    ranges = [[-5, 5], [-5, 5], [-0.2, 0.2]]
    resolution = 0.1

    countshape = [numpy.round((r[1] - r[0]) / resolution).astype('i4') + 2
                  for r in ranges]

    coords = [numpy.clip(numpy.floor((c - r[0]) / resolution).astype('i4'),
                         -1, s - 2) + 1
              for (c, r, s) in zip(xyz, ranges, countshape)]
    centers = [r[0] + resolution * (numpy.arange((r[1] - r[0]) / resolution) + 0.5)
               for r in ranges]

    flatcoord = numpy.ravel_multi_index(coords, countshape)

    countsd = numpy.bincount(flatcoord[dustsel & ~othersel],
                             minlength=numpy.product(countshape))
    countso = numpy.bincount(flatcoord[othersel & dustsel],
                             minlength=numpy.product(countshape))
    countsd = countsd.reshape(countshape)
    countso = countso.reshape(countshape)

    pnumdensd = map_coordinates_wrap((countsd[1:-1, 1:-1, 1:-1], centers),
                                     [c[dustsel & ~othersel] for c in xyz],
                                     order=0)
    pnumdenso = map_coordinates_wrap((countso[1:-1, 1:-1, 1:-1], centers),
                                     [c[dustsel & ~othersel] for c in xyz],
                                     order=0)

    ntopick = 10. / downsampledby
    pnumdensleft = ntopick - pnumdenso

    numpy.random.seed(0)
    randnum = numpy.random.rand(numpy.sum(dustsel & ~othersel))

    subsel = (randnum < pnumdensleft / 1. / (pnumdensd + (pnumdensd == 0)))

    dustsel[othersel] = 0
    dustsel[dustsel & ~othersel] = subsel

    return dustsel


def map_coordinates_wrap(grid, coord, **kw):
    gridpts = grid[0]
    gridcoord = grid[1]
    outputcoordnorm = [numpy.interp(c, gc, numpy.arange(len(gc)))
                       for c, gc in zip(coord, gridcoord)]
    from scipy.ndimage import interpolation
    return interpolation.map_coordinates(gridpts, outputcoordnorm,
                                         cval=0., mode='constant', **kw)


# These were initial selection criteria, not used anymore. We use the
# real GG carton instead.

def jkselect(data):
    jk0 = data['j_ks_0']
    return (jk0 > 0.7) & (data['hmag'] < 11)


def ghselect(data):
    gaiag = data['gaiamag']
    return ((data['hmag'] < 11) & numpy.where(numpy.isfinite(gaiag),
                                              gaiag - data['hmag'] > 3.5,
                                              True))


class MWM_Dust_Carton(BaseCarton):
    """MWM Dust Carton.

    Definition:

        A sample of bright, nearby, midplane giants that, when combined
        with the more distant and/or heavily reddened sample in other cartons,
        produces 100 stars per (100pc)^3 volume in the near disk.

        - Use Gaia parallax and (l,b) to get distance
        - Use H and 4.5 micron to get A_Ks and E_JKs, a la APOGEE
          (using RJCE dereddening prescription from Majewski et al. 2011).
        - Use J, K, and E_JKs to get (J-Ks)_0
        - Use distance, K, and A_Ks to get absolute mag M_K
        - 0 < Gaia parallax error/parallax < 0.2
        - M_K < 2.6
        - H < 11.2
        - distance < 5 kpc
        - |z| < 0.2 kpc
        - (J-Ks)_0 > 0.5
        - Can be spatially subselected to complement GG sampling in order to
          obtain 100 stars per (100pc)^3 volume.
        - Galactic Genesis quality flags:
            (ph_qual[1] = 'A' OR ph_qual[1] = 'B')
                AND gal_contam=0
                AND cc_flg[1]=0
                AND (rd_flg[1] > 0 AND rd_flg[1] <= 3)

    Non-SQL implementation:

        https://faun.rc.fas.harvard.edu/eschlafly/sdss5/dustsel.py

    """

    name = 'mwm_dust_core'
    mapper = 'MWM'
    category = 'science'
    program = 'mwm_dust'
    cadence = 'mwm_dust_1x1'
    priority = 2720

    def build_query(self, version_id, query_region=None):

        # Do a quick check to be sure the GG carton exists in targetdb.
        gg_exists = (targetdb.Carton.select()
                     .join(targetdb.Version)
                     .where(targetdb.Carton.carton == 'mwm_galactic',
                            targetdb.Version.plan == self.plan,
                            targetdb.Version.target_selection >> True)
                     .exists())
        if not gg_exists:
            raise RuntimeError('mwm_galactic has not been loaded yet.')

        fn = peewee.fn

        # GG quality flags
        ph_qual = TwoMassPSC.ph_qual
        cc_flg = TwoMassPSC.cc_flg
        rd_flg = TwoMassPSC.rd_flg
        rd_flag_1 = peewee.fn.substr(rd_flg, 2, 1).cast('integer')
        gal_contam = TwoMassPSC.gal_contam

        gallong = TIC_v8.gallong
        gallat = TIC_v8.gallat

        ipar = 1. / TIC_v8.plx  # kpc
        zz = ipar * fn.sin(fn.radians(gallat))
        xx = ipar * fn.cos(fn.radians(gallat)) * fn.cos(fn.radians(gallong))
        yy = ipar * fn.cos(fn.radians(gallat)) * fn.sin(fn.radians(gallong))

        dist = fn.sqrt(fn.pow(xx, 2) + fn.pow(yy, 2) + fn.pow(zz, 2))

        aks_glimpse = 0.918 * (GLIMPSE.mag_h - GLIMPSE.mag4_5 - 0.08)
        aks_allwise = 0.918 * (AllWise.h_m_2mass - AllWise.w2mpro - 0.08)
        aks = fn.coalesce(aks_glimpse, aks_allwise)

        Ej_ks = 1.5 * aks

        j_ks_0_glimpse = GLIMPSE.mag_j - GLIMPSE.mag_ks - Ej_ks
        j_ks_0_allwise = AllWise.j_m_2mass - AllWise.k_m_2mass - Ej_ks
        j_ks_0 = fn.coalesce(j_ks_0_glimpse, j_ks_0_allwise)

        plxfracunc = TIC_v8.e_plx / TIC_v8.plx
        dm = 5 * fn.log(1000. / TIC_v8.plx / 10)
        absmag = TIC_v8.kmag - aks - dm

        query = (CatalogToTIC_v8
                 .select(CatalogToTIC_v8.catalogid,
                         peewee.Value(False).alias('selected'),  # Set selected to False
                         TIC_v8.gaia_int.alias('gaia_souce_id'),
                         TIC_v8.gallong, TIC_v8.gallat,
                         TIC_v8.plx,
                         TIC_v8.gaiamag,
                         TIC_v8.hmag,
                         TIC_v8.kmag,
                         aks.alias('a_ks'),
                         j_ks_0.alias('j_ks_0'))
                 .join(TIC_v8)
                 .join_from(CatalogToTIC_v8, CatalogToAllWise,
                            peewee.JOIN.LEFT_OUTER,
                            on=(CatalogToAllWise.catalogid == CatalogToTIC_v8.catalogid))
                 .join(AllWise, peewee.JOIN.LEFT_OUTER)
                 .join_from(TIC_v8, TwoMassPSC, peewee.JOIN.LEFT_OUTER)
                 .join_from(CatalogToAllWise, CatalogToGLIMPSE,
                            peewee.JOIN.LEFT_OUTER,
                            on=(CatalogToGLIMPSE.catalogid == CatalogToAllWise.catalogid))
                 .join(GLIMPSE, peewee.JOIN.LEFT_OUTER)
                 .where(((CatalogToAllWise.version_id == version_id) &
                         (CatalogToAllWise.best >> True)) |
                        (CatalogToAllWise.catalogid >> None))
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True)
                 .where(((CatalogToGLIMPSE.version_id == version_id) &
                         (CatalogToGLIMPSE.best >> True)) |
                        (CatalogToGLIMPSE.catalogid >> None))
                 .where(TIC_v8.hmag < 11.2,
                        fn.abs(zz) < 0.2,
                        j_ks_0.is_null(False), j_ks_0 > 0.5,
                        dist < 5,
                        plxfracunc < 0.2, plxfracunc > 0,
                        absmag < 2.6)
                 .where(ph_qual.regexp('.(A|B).'),
                        gal_contam == 0,
                        peewee.fn.substr(cc_flg, 2, 1) == '0',
                        rd_flag_1 > 0, rd_flag_1 <= 3))

        if query_region:
            query = (query
                     .join_from(CatalogToAllWise, Catalog)
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))

        return query

    def post_process(self, model):
        """Select samples based on GG using Eddie Schlafly code.

        Goal subsample and to select only targets that are not already part
        of the Galactic Genesis sample. We use the real GG sample from
        targetdb.

        """

        data = pandas.read_sql(
            f'SELECT * FROM {model._meta.schema}.{model._meta.table_name};',
            self.database)

        mwm_galactic = (targetdb.Target
                        .select(targetdb.Target.catalogid)
                        .join(targetdb.CartonToTarget)
                        .join(targetdb.Carton)
                        .join(targetdb.Version)
                        .where(targetdb.Carton.carton == 'mwm_galactic',
                               targetdb.Version.plan == self.plan,
                               targetdb.Version.target_selection >> True))

        gg_catids = tuple(zip(*mwm_galactic.tuples()))[0]
        gg_mask = numpy.in1d(data.catalogid, gg_catids)

        # Subsample based on GG and update DB.
        dust_gg_subsel = subselect(data, gg_mask)
        dust_gg_cid = peewee.ValuesList(zip(data.catalogid[dust_gg_subsel]),
                                        columns=('catalogid',), alias='vl')

        (model
         .update({model.selected: True})
         .from_(dust_gg_cid)
         .where(model.catalogid == dust_gg_cid.c.catalogid)
         .execute())

        return
